from __future__ import annotations

from unittest import mock

import zocalo.cli.dlq_check
from zocalo.configuration import Configuration


@mock.patch("zocalo.cli.dlq_check.JMXAPI")
def test_activemq_dlq_check(mock_jmx):
    cfg = Configuration({})
    _mock = mock.Mock()
    mock_jmx.return_value = _mock
    mock_jmx.return_value.org.apache.activemq.return_value = {
        "status": 200,
        "value": {
            "destinationName=images": {"QueueSize": 2},
            "destinationName=transient": {"QueueSize": 5},
        },
    }
    checked = zocalo.cli.dlq_check.check_dlq(cfg, "zocalo")
    assert checked == {"images": 2, "transient": 5}


def test_activemq_dlq_rabbitmq_check(requests_mock):
    zc = mock.Mock()
    zc.rabbitmqapi = {
        "base_url": "http://fake.com/api",
        "username": "guest",
        "password": "guest",
    }
    requests_mock.get(
        "/api/queues",
        json=[
            {"name": "images", "vhost": "zocalo", "messages": 10, "exclusive": False},
            {
                "name": "dlq.images",
                "vhost": "zocalo",
                "messages": 2,
                "exclusive": False,
            },
            {
                "name": "dlq.transient",
                "vhost": "zocalo",
                "messages": 5,
                "exclusive": False,
            },
        ],
    )
    requests_mock.get("/api/health/checks/alarms", json={"status": "ok"})

    checked = zocalo.cli.dlq_check.check_dlq_rabbitmq(zc, "zocalo")
    assert checked == {"dlq.images": 2, "dlq.transient": 5}



#
# zocalo.dlq_purge
#   Retrieve all dead letter queue messages from ActiveMQ and store them
#   in a temporary directory.
#

from __future__ import annotations

import argparse
import json
import pathlib
import queue
import re
import sys
import time
from datetime import datetime
from functools import partial

import workflows

import zocalo.configuration
from zocalo.util.rabbitmq import RabbitMQAPI


def run() -> None:
    zc = zocalo.configuration.from_file()
    zc.activate()
    parser = argparse.ArgumentParser(
        usage="zocalo.dlq_purge [options] [queue [queue ...]]"
    )

    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    dlqprefix = "zocalo"
    parser.add_argument(
        "--wait",
        action="store",
        dest="wait",
        type=float,
        help="Wait this many seconds for incoming messages",
    )
    if zc.storage and zc.storage.get("zocalo.dlq.purge_location"):
        dlq_dump_path = zc.storage["zocalo.dlq.purge_location"]
    else:
        dlq_dump_path = "./DLQ"
    parser.add_argument(
        "--location",
        action="store",
        dest="location",
        type=str,
        default=dlq_dump_path,
        help=f"Where to write out DLQ message files (default: {dlq_dump_path})",
    )
    parser.add_argument(
        "queues",
        nargs="*",
        help="Queues to purge of dead letters. For RabbitMQ do not include the dlq. prefix in the queue names",
    )
    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)

    args = parser.parse_args(["--stomp-prfx=DLQ"] + sys.argv[1:])
    if args.transport == "PikaTransport":
        queues = ["dlq." + a for a in args.queues]
    else:
        queues = args.queues
    transport = workflows.transport.lookup(args.transport)()

    characterfilter = re.compile(r"[^a-zA-Z0-9._-]+", re.UNICODE)
    idlequeue: queue.Queue = queue.Queue()

    def receive_dlq_message(header: dict, message: dict, rabbitmq=False) -> None:
        idlequeue.put_nowait("start")
        if rabbitmq:
            msg_time = int(datetime.timestamp(header["x-death"][0]["time"])) * 1000
            header["x-death"][0]["time"] = datetime.timestamp(
                header["x-death"][0]["time"]
            )
        else:
            msg_time = int(header["timestamp"])
        timestamp = time.localtime(msg_time / 1000)
        millisec = msg_time % 1000
        filepath = pathlib.Path(args.location, time.strftime("%Y-%m-%d", timestamp))
        filepath.mkdir(parents=True, exist_ok=True)
        filename = filepath / (
            "msg-"
            + time.strftime("%Y%m%d-%H%M%S", timestamp)
            + f"-{millisec:03d}-"
            + characterfilter.sub("_", str(header["message-id"]))
        )

        dlqmsg = {
            "exported": {
                "date": time.strftime("%Y-%m-%d"),
                "time": time.strftime("%H:%M:%S"),
            },
            "header": header,
            "message": message,
        }

        with filename.open("w") as fh:
            json.dump(dlqmsg, fh, indent=2, sort_keys=True)
        print(
            f"Message {header['message-id']} ({time.strftime('%Y-%m-%d %H:%M:%S', timestamp)}) exported:\n  {filename}"
        )
        transport.ack(header)
        idlequeue.put_nowait("done")

    transport.connect()
    if not queues:
        if args.transport == "StompTransport":
            queues = [dlqprefix + ".>"]
        elif args.transport == "PikaTransport":
            rmq = RabbitMQAPI.from_zocalo_configuration(zc)
            queues = [q.name for q in rmq.queues() if q.name.startswith("dlq.")]
    for queue_ in queues:
        print("Looking for DLQ messages in " + queue_)
        transport.subscribe(
            queue_,
            partial(receive_dlq_message, rabbitmq=args.transport == "PikaTransport"),
            acknowledgement=True,
        )
    try:
        idlequeue.get(True, args.wait or 3)
        while True:
            idlequeue.get(True, args.wait or 0.1)
    except queue.Empty:
        print("Done.")
    transport.disconnect()



from __future__ import annotations

import re

import pytest

import zocalo.configuration
import zocalo.util.rabbitmq as rabbitmq


@pytest.fixture
def zocalo_configuration(mocker):
    zc = mocker.MagicMock(zocalo.configuration.Configuration)
    zc.rabbitmqapi = {
        "base_url": "http://rabbitmq.burrow.com:12345/api",
        "username": "carrots",
        "password": "carrots",
    }
    return zc


@pytest.fixture
def rmqapi(requests_mock, zocalo_configuration):
    requests_mock.get(re.compile("/health/checks/alarms"), json={"status": "ok"})
    api = rabbitmq.RabbitMQAPI.from_zocalo_configuration(zocalo_configuration)
    requests_mock.reset_mock()
    return api


def test_http_api_request(zocalo_configuration):
    request = rabbitmq.http_api_request(zocalo_configuration, api_path="/queues")
    assert request.get_full_url() == "http://rabbitmq.burrow.com:12345/api/queues"


def test_api_health_checks(requests_mock, rmqapi):
    requests_mock.get(re.compile("/health/checks/"), json={"status": "ok"})
    success, failures = rmqapi.health_checks
    assert not failures
    assert success
    for k, v in success.items():
        assert k.startswith("health/checks/")
        assert v == {"status": "ok"}


def test_api_health_checks_failures(requests_mock, rmqapi):
    expected_json = {
        "status": "failed",
        "reason": "No active listener",
        "missing": 1234,
        "ports": [25672, 15672, 1883, 15692, 61613, 5672],
    }
    requests_mock.get(re.compile("/health/checks/"), json={"status": "ok"})
    requests_mock.get(
        re.compile("/health/checks/port-listener"),
        status_code=503,
        reason="No active listener",
        json=expected_json,
    )
    success, failures = rmqapi.health_checks
    assert failures
    assert success
    assert len(failures) == 1
    for k, v in success.items():
        assert k.startswith("health/checks/")
        assert v == {"status": "ok"}
    for k, v in failures.items():
        assert k.startswith("health/checks/port-listener/")
        assert v == expected_json


def test_api_queues(requests_mock, rmqapi):
    queue = {
        "consumers": 0,
        "exclusive": False,
        "memory": 110112,
        "message_stats": {
            "deliver_get": 33,
            "deliver_get_details": {"rate": 0},
            "publish": 22,
            "publish_details": {"rate": 0},
        },
        "messages": 0,
        "messages_ready": 0,
        "messages_unacknowledged": 0,
        "name": "foo",
        "vhost": "zocalo",
    }

    # First call rmq.queues() with defaults
    requests_mock.get("/api/queues", json=[queue])
    assert rmqapi.queues() == [rabbitmq.QueueInfo(**queue)]

    # Now call with vhost=...
    requests_mock.get("/api/queues/zocalo", json=[queue])
    assert rmqapi.queues(vhost="zocalo") == [rabbitmq.QueueInfo(**queue)]

    # Now call with vhost=..., name=...
    requests_mock.get(f"/api/queues/zocalo/{queue['name']}", json=queue)
    assert rmqapi.queues(vhost="zocalo", name=queue["name"]) == rabbitmq.QueueInfo(
        **queue
    )


@pytest.fixture
def queue_spec():
    return rabbitmq.QueueSpec(
        name="foo",
        auto_delete=True,
        vhost="zocalo",
        arguments={"x-single-active-consumer": True},
    )


def test_api_queue_declare(requests_mock, rmqapi, queue_spec):
    requests_mock.put("/api/queues/zocalo/foo")
    rmqapi.queue_declare(queue=queue_spec)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "PUT"
    assert history.url.endswith("/api/queues/zocalo/foo")
    assert history.json() == {
        "auto_delete": True,
        "arguments": queue_spec.arguments,
    }


def test_api_queue_delete(requests_mock, rmqapi, queue_spec):
    requests_mock.delete("/api/queues/zocalo/foo")
    rmqapi.queue_delete(vhost="zocalo", name="foo", if_unused=True, if_empty=True)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "DELETE"
    assert history.url.endswith("/api/queues/zocalo/foo?if-unused=True&if-empty=True")


def test_api_bindings(requests_mock, rmqapi):
    binding = {
        "source": "foo",
        "destination": "bar",
        "destination_type": "q",
        "arguments": {},
        "routing_key": "bar",
        "properties_key": "bar",
        "vhost": "zocalo",
    }
    response = {
        "source": "foo",
        "destination": "bar",
        "destination_type": "queue",
        "arguments": {},
        "routing_key": "bar",
        "properties_key": "bar",
        "vhost": "zocalo",
    }

    requests_mock.get("/api/bindings", json=[response])
    assert rmqapi.bindings() == [rabbitmq.BindingInfo(**binding)]

    requests_mock.get("/api/bindings/zocalo", json=[response])
    assert rmqapi.bindings(vhost="zocalo") == [rabbitmq.BindingInfo(**binding)]

    requests_mock.get(
        f"/api/bindings/zocalo/e/{binding['source']}/q/{binding['destination']}",
        json=[response],
    )
    assert rmqapi.bindings(
        vhost="zocalo",
        source=binding["source"],
        destination=binding["destination"],
        destination_type="q",
    ) == [rabbitmq.BindingInfo(**binding)]


@pytest.fixture
def binding_spec():
    return rabbitmq.BindingSpec(
        source="foo",
        destination="bar",
        destination_type="q",
        arguments={},
        routing_key="bar",
        vhost="zocalo",
    )


def test_api_binding_declare(requests_mock, rmqapi, binding_spec):
    requests_mock.post("/api/bindings/zocalo/e/foo/q/bar")
    rmqapi.binding_declare(binding=binding_spec)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "POST"
    assert history.url.endswith("/api/bindings/zocalo/e/foo/q/bar")
    assert history.json() == {
        "arguments": binding_spec.arguments,
        "routing_key": "bar",
    }


def test_api_bindings_delete(requests_mock, rmqapi, binding_spec):
    binding = {
        "source": "foo",
        "destination": "bar",
        "destination_type": "queue",
        "arguments": {},
        "routing_key": "bar",
        "properties_key": "bar",
        "vhost": "zocalo",
    }
    requests_mock.get("/api/bindings/zocalo/e/foo/q/bar", json=[binding])
    requests_mock.delete("/api/bindings/zocalo/e/foo/q/bar/bar")
    rmqapi.bindings_delete(
        vhost="zocalo",
        source="foo",
        destination="bar",
        destination_type="q",
    )
    assert requests_mock.call_count == 2
    history = requests_mock.request_history[0]
    assert history.method == "GET"
    assert history.url.endswith("/api/bindings/zocalo/e/foo/q/bar")
    history = requests_mock.request_history[1]
    assert history.method == "DELETE"
    assert history.url.endswith("/api/bindings/zocalo/e/foo/q/bar/bar")


def test_api_nodes(requests_mock, rmqapi):
    node = {
        "name": "rabbit@pooter123",
        "mem_limit": 80861855744,
        "mem_alarm": False,
        "mem_used": 143544320,
        "disk_free_limit": 50000000,
        "disk_free_alarm": False,
        "disk_free": 875837644800,
        "fd_total": 32768,
        "fd_used": 56,
        "io_file_handle_open_attempt_count": 647,
        "sockets_total": 29401,
        "sockets_used": 0,
        "gc_num": 153378077,
        "gc_bytes_reclaimed": 7998215046336,
        "proc_total": 1048576,
        "proc_used": 590,
        "run_queue": 1,
        "running": True,
        "type": "disc",
    }

    # First call rmq.nodes() with defaults
    requests_mock.get("/api/nodes", json=[node])
    assert rmqapi.nodes() == [rabbitmq.NodeInfo(**node)]

    # Now call with name=...
    requests_mock.get(f"/api/nodes/{node['name']}", json=node)
    assert rmqapi.nodes(name=node["name"]) == rabbitmq.NodeInfo(**node)


@pytest.mark.parametrize("name", ["", "foo"])
def test_api_exchanges(name, requests_mock, rmqapi):
    exchange = {
        "arguments": {},
        "auto_delete": False,
        "durable": True,
        "internal": False,
        "message_stats": {
            "publish_in": 156447,
            "publish_in_details": {"rate": 0.4},
            "publish_out": 156445,
            "publish_out_details": {"rate": 0.4},
        },
        "name": name,
        "type": "direct",
        "user_who_performed_action": "rmq-internal",
        "vhost": "foo",
    }

    # First call rmq.exchanges() with defaults
    requests_mock.get("/api/exchanges", json=[exchange])
    assert rmqapi.exchanges() == [rabbitmq.ExchangeInfo(**exchange)]

    # Now call with vhost=...
    requests_mock.get("/api/exchanges/zocalo/", json=[exchange])
    assert rmqapi.exchanges(vhost="zocalo") == [rabbitmq.ExchangeInfo(**exchange)]

    # Now call with vhost=..., name=...
    requests_mock.get(
        f"/api/exchanges/{exchange['vhost']}/{exchange['name']}/", json=exchange
    )
    assert rmqapi.exchanges(
        vhost=exchange["vhost"], name=exchange["name"]
    ) == rabbitmq.ExchangeInfo(**exchange)


def exchange_spec(name):
    return rabbitmq.ExchangeSpec(
        name=name,
        type="fanout",
        durable=True,
        auto_delete=True,
        internal=False,
        vhost="zocalo",
    )


@pytest.mark.parametrize("name", ["", "foo"])
def test_api_exchange_declare(name, requests_mock, rmqapi):
    requests_mock.put(f"/api/exchanges/zocalo/{name}/")
    rmqapi.exchange_declare(exchange=exchange_spec(name))
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "PUT"
    assert history.url.endswith(f"/api/exchanges/zocalo/{name}/")
    assert history.json() == {
        "type": "fanout",
        "auto_delete": True,
        "durable": True,
        "arguments": {},
    }


def test_api_exchange_delete(requests_mock, rmqapi):
    requests_mock.delete("/api/exchanges/zocalo/foo")
    rmqapi.exchange_delete(vhost="zocalo", name="foo", if_unused=True)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "DELETE"
    assert history.url.endswith("/api/exchanges/zocalo/foo?if-unused=True")


def test_api_connections(requests_mock, rmqapi):
    connection = {
        "auth_mechanism": "PLAIN",
        "connected_at": 1634716019864,
        "frame_max": 131072,
        "host": "123.24.5.67",
        "name": "123.24.5.67:12345 -> 123.24.5.67:54321",
        "node": "rabbit@cs05r-sc-serv-26",
        "peer_host": "123.24.5.67",
        "peer_port": 12345,
        "port": 54321,
        "protocol": "AMQP 0-9-1",
        "ssl": False,
        "state": "running",
        "timeout": 60,
        "user": "foo",
        "vhost": "bar",
        "channels": 1,
    }

    # First call rmq.connections() with defaults
    requests_mock.get("/api/connections", json=[connection])
    assert rmqapi.connections() == [rabbitmq.ConnectionInfo(**connection)]

    # Now call with name=...
    requests_mock.get(f"/api/connections/{connection['name']}/", json=connection)
    assert rmqapi.connections(name=connection["name"]) == rabbitmq.ConnectionInfo(
        **connection
    )


def test_api_users(requests_mock, rmqapi):
    user = {
        "name": "guest",
        "password_hash": "guest",
        "hashing_algorithm": "rabbit_password_hashing_sha256",
        "tags": ["administrator"],
    }

    # First call rmq.users() with defaults
    requests_mock.get("/api/users", json=[user])
    assert rmqapi.users() == [rabbitmq.UserSpec(**user)]

    # Now call with name=...
    requests_mock.get(f"/api/users/{user['name']}/", json=user)
    assert rmqapi.user(user["name"]) == rabbitmq.UserSpec(**user)


@pytest.fixture
def user_spec():
    return rabbitmq.UserSpec(
        name="guest",
        password_hash="guest",
        hashing_algorithm="rabbit_password_hashing_sha256",
        tags=["administrator"],
    )


def test_api_add_user(requests_mock, rmqapi, user_spec):
    requests_mock.put(f"/api/users/{user_spec.name}/")
    rmqapi.user_put(user=user_spec)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "PUT"
    assert history.url.endswith(f"/api/users/{user_spec.name}/")
    assert history.json() == {
        "password_hash": "guest",
        "hashing_algorithm": "rabbit_password_hashing_sha256",
        "tags": "administrator",
    }


def test_api_delete_user(requests_mock, rmqapi, user_spec):
    requests_mock.delete("/api/users/guest/")
    rmqapi.user_delete(name="guest")
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "DELETE"
    assert history.url.endswith("/api/users/guest/")


def test_api_permissions(requests_mock, rmqapi):
    permissions = [
        {
            "vhost": "my-vhost",
            "user": "janeway",
            "configure": "^janeway-.*",
            "write": ".*",
            "read": ".*",
        },
        {
            "vhost": "my-vhost",
            "user": "admin",
            "configure": "^.*",
            "write": ".*",
            "read": ".*",
        },
    ]

    # First call rmq.permissions() with defaults
    requests_mock.get("/api/permissions", json=permissions)
    assert rmqapi.permissions() == [rabbitmq.PermissionSpec(**p) for p in permissions]

    # Now call with name=...
    requests_mock.get("/api/permissions/my-vhost/janeway/", json=permissions[0])
    assert rmqapi.permissions(
        vhost="my-vhost", user="janeway"
    ) == rabbitmq.PermissionSpec(**permissions[0])


@pytest.fixture
def permission_spec():
    return rabbitmq.PermissionSpec(
        vhost="my-vhost",
        user="janeway",
        configure="^janeway-.*",
        write=".*",
        read=".*",
    )


def test_api_add_permissions(requests_mock, rmqapi, permission_spec):
    endpoint = f"/api/permissions/{permission_spec.vhost}/{permission_spec.user}/"
    requests_mock.put(endpoint)
    rmqapi.set_permissions(permission_spec)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "PUT"
    assert history.url.endswith(endpoint)
    assert history.json() == {
        "configure": "^janeway-.*",
        "write": ".*",
        "read": ".*",
    }


def test_api_clear_permissions(requests_mock, rmqapi):
    endpoint = "/api/permissions/foo/bar/"
    requests_mock.delete(endpoint)
    rmqapi.clear_permissions(vhost="foo", user="bar")
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "DELETE"
    assert history.url.endswith(endpoint)


def test_api_vhosts(requests_mock, rmqapi):
    vhosts = [
        {
            "name": "foo",
            "description": "",
            "tags": [],
            "tracing": False,
        },
        {
            "name": "bar",
            "description": "This is a description",
            "tags": ["ham", "spam"],
            "tracing": True,
        },
    ]

    # First call rmq.users() with defaults
    requests_mock.get("/api/vhosts", json=vhosts)
    assert rmqapi.vhosts() == [rabbitmq.VHostSpec(**vhost) for vhost in vhosts]

    # Now call with name=...
    requests_mock.get("/api/vhosts/foo/", json=vhosts[0])
    assert rmqapi.vhost("foo") == rabbitmq.VHostSpec(**vhosts[0])


@pytest.fixture
def vhost_spec():
    return rabbitmq.VHostSpec(
        name="foo",
    )


def test_api_add_vhost(requests_mock, rmqapi, vhost_spec):
    requests_mock.put(f"/api/vhosts/{vhost_spec.name}/")
    rmqapi.add_vhost(vhost=vhost_spec)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "PUT"
    assert history.url.endswith(f"/api/vhosts/{vhost_spec.name}/")
    assert history.json() == {
        "tags": [],
    }


def test_api_delete_vhost(requests_mock, rmqapi, vhost_spec):
    requests_mock.delete(f"/api/vhosts/{vhost_spec.name}/")
    rmqapi.delete_vhost(name=vhost_spec.name)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "DELETE"
    assert history.url.endswith(f"/api/vhosts/{vhost_spec.name}/")


def test_api_policies(requests_mock, rmqapi):
    policy = {
        "vhost": "foo",
        "name": "redelivery",
        "pattern": "^amq.",
        "apply-to": "queues",
        "definition": {"delivery-limit": 5},
        "priority": 0,
    }

    # First call rmq.policies() with defaults
    requests_mock.get("/api/policies", json=[policy])
    assert rmqapi.policies() == [rabbitmq.PolicySpec(**policy)]

    # Now call with vhost=...
    requests_mock.get(f"/api/policies/{policy['vhost']}/", json=[policy])
    assert rmqapi.policies(vhost=policy["vhost"]) == [rabbitmq.PolicySpec(**policy)]

    # Now call with vhost=..., name=...
    requests_mock.get(f"/api/policies/{policy['vhost']}/{policy['name']}/", json=policy)
    assert rmqapi.policy(
        vhost=policy["vhost"], name=policy["name"]
    ) == rabbitmq.PolicySpec(**policy)


@pytest.fixture
def policy_spec():
    return rabbitmq.PolicySpec(
        name="bar",
        pattern="^amq.",
        apply_to=rabbitmq.PolicyApplyTo.queues,
        definition={"delivery-limit": 5},
        vhost="foo",
    )


def test_api_set_policy(requests_mock, rmqapi, policy_spec):
    requests_mock.put(f"/api/policies/foo/{policy_spec.name}/")
    rmqapi.set_policy(policy=policy_spec)
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "PUT"
    assert history.url.endswith(f"/api/policies/foo/{policy_spec.name}/")
    assert history.json() == {
        "pattern": "^amq.",
        "apply-to": "queues",
        "definition": {"delivery-limit": 5},
    }


def test_api_clear_policy(requests_mock, rmqapi, policy_spec):
    requests_mock.delete("/api/policies/foo/bar/")
    rmqapi.clear_policy(vhost="foo", name="bar")
    assert requests_mock.call_count == 1
    history = requests_mock.request_history[0]
    assert history.method == "DELETE"
    assert history.url.endswith("/api/policies/foo/bar/")



#
# zocalo.shutdown
#   Stop a zocalo service
#

from __future__ import annotations

import argparse
import socket
import sys

import workflows
import workflows.services
import workflows.transport

import zocalo.configuration


def run(args=None):
    parser = argparse.ArgumentParser()

    # Load configuration
    zc = zocalo.configuration.from_file()
    zc.activate()

    known_services = sorted(workflows.services.get_known_services())
    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "HOSTS",
        nargs="*",
        type=str,
        help="Specific service instances specified as hostname.pid",
    )
    parser.add_argument(
        "-s",
        "--service",
        dest="services",
        metavar="SVC",
        action="append",
        default=[],
        help="Stop all instances of a service. Use 'none' for instances without "
        "loaded service. Known services: " + ", ".join(known_services),
    )
    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)
    args = parser.parse_args(args)

    if not args.services and not len(args.HOSTS):
        print("Need to specify one or more services to shut down.")
        print("Either specify service groups with -s or specify specific instances")
        print("as: hostname.pid")
        sys.exit(1)

    transport = workflows.transport.lookup(args.transport)()
    transport.connect()

    for host in args.HOSTS:
        if host.count(".") == 1:
            # See also workflows.util.generate_unique_host_id()
            host = ".".join(reversed(socket.gethostname().split(".")[1:])) + "." + host

        message = {"command": "shutdown", "host": host}
        transport.broadcast("command", message)
        print("Shutting down", host)

    for service in args.services:
        if service.lower() == "none":
            # Special case for placeholder instances
            service = None
        message = {"command": "shutdown", "service": service}
        transport.broadcast("command", message)
        print("Stopping all instances of", service)

    transport.disconnect()



#
# simple API to obtain JMX information
#
# Point to a Zocalo configuration object with a loaded jmx plugin instance to use it, eg:
#  zc = zocalo.configuration.from_string("...")
#  zc.activate_environment("environment-with-a-jmx-plugin-configuration")
#  jmx = JMXAPI(zc)
# Then you can access objects with eg.
#  jmx.java.lang(type="Memory")
#  jmx.org.apache.activemq(type="Broker", brokerName="localhost/TotalConsumerCount")


from __future__ import annotations

import base64
import json
import urllib.request

import zocalo.configuration


class JMXAPIPath:
    """A recursing helper object that encodes a JMX bean path."""

    def __init__(self, path, callback):
        self.path = path
        self.callback = callback

    def __repr__(self):
        return self.path

    def __getattribute__(self, attribute):
        try:
            return object.__getattribute__(self, attribute)
        except AttributeError:
            return JMXAPIPath(self.path + "." + attribute, self.callback)

    def __call__(self, *args, **kwargs):
        return self.callback(self.path, *args, **kwargs)


class JMXAPI:
    """Access to JMX via the Joloika/REST API to obtain monitoring information
    from a running JVM."""

    def __init__(self, zc: zocalo.configuration.Configuration):
        if not zc.jmx:
            raise zocalo.ConfigurationError(
                "There are no JMX credentials configured in your environment"
            )
        self.url = (
            f"http://{zc.jmx['host']}:{zc.jmx['port']}/{zc.jmx['base_url']}/read/"
        )
        self._authstring = b"Basic " + base64.b64encode(
            zc.jmx["username"].encode("utf-8")
            + b":"
            + zc.jmx["password"].encode("utf-8")
        )

    def __getattribute__(self, attribute):
        try:
            return object.__getattribute__(self, attribute)
        except AttributeError:
            return JMXAPIPath(attribute, self._call)

    def _call(self, path, attribute=None, *args, **kwargs):
        params = ",".join(key + "=" + value for key, value in kwargs.items())
        url = path + ":" + params
        if attribute:
            url = url + "/" + attribute
        return self._get(url)

    def _get(self, url):
        complete_url = self.url + url
        req = urllib.request.Request(
            complete_url, headers={"Accept": "application/json"}
        )
        req.add_header("Authorization", self._authstring)
        handler = urllib.request.urlopen(req)
        returncode = handler.getcode()
        if returncode != 200:
            raise RuntimeError("JMX lookup returned HTTP code %d" % returncode)
        return json.load(handler)


if __name__ == "__main__":
    zc = zocalo.configuration.from_file()
    if "live" in zc.environments:
        zc.activate_environment("live")
    jmx = JMXAPI(zc)
    from pprint import pprint

    mem = jmx.java.lang(type="Memory")
    pprint(mem)
    consumers = jmx.org.apache.activemq(
        type="Broker", brokerName="localhost", attribute="TotalConsumerCount"
    )
    pprint(consumers)
    health = jmx.org.apache.activemq(
        type="Broker", brokerName="localhost", service="Health"
    )
    pprint(health)
    queuestats = jmx.org.apache.activemq(
        type="Broker",
        brokerName="localhost",
        destinationType="Queue",
        destinationName="zocalo.transient.controller",
        attribute="QueueSize",
    )
    pprint(queuestats)



from __future__ import annotations

import copy
import logging.config
import socket
from typing import Any

import graypy.handler

import zocalo


def _config_is_incremental(config: dict[str, Any]) -> bool:
    forbidden_keys = {"filters", "formatters", "handlers"}

    # Incremental configurations must not declare any of those
    if any(key in config for key in forbidden_keys):
        return False

    # Incremental configurations must not attach any of those
    # to the root logger
    if any(key in config.get("root", {}) for key in forbidden_keys):
        return False
    # or any other logger
    for logger in config.get("loggers", {}).values():
        if any(key in logger for key in forbidden_keys):
            return False

    return True


class LoggingIncrementer:
    def __init__(self, setup: dict[str, Any]):
        self._setup = setup
        self._verbosity_level = 0

    @property
    def verbosity(self) -> int:
        return self._verbosity_level

    @verbosity.setter
    def verbosity(self, value: int) -> None:
        if value <= self._verbosity_level:
            return
        for verbosity_level in range(self._verbosity_level, value):
            try:
                incremental = self._setup.get("verbose", [])[verbosity_level]
            except IndexError:
                break
            logging.config.dictConfig(incremental)
            self._verbosity_level = verbosity_level + 1

    def __repr__(self) -> str:
        return f"<LoggingConfiguration verbosity={self._verbosity_level}>"


class Logging:
    """
    A plugin to set up consistent logging configuration
    using a Zocalo configuration file.
    """

    @staticmethod
    def activate(configuration: dict) -> LoggingIncrementer:
        logconfig = copy.deepcopy(configuration)

        del logconfig["plugin"]
        logconfig.setdefault("version", 1)
        logconfig.setdefault("disable_existing_loggers", False)
        logconfig.setdefault("incremental", False)

        if logconfig["incremental"] and not _config_is_incremental(logconfig):
            raise zocalo.ConfigurationError(
                "Logging configuration error: definition defines items not allowed "
                "in an incremental definition"
            )

        logging.config.dictConfig(logconfig)

        for level, verbosity_def in enumerate(logconfig.get("verbose", [])):
            if not isinstance(verbosity_def, dict):
                raise zocalo.ConfigurationError(
                    f"Logging configuration error: verbosity level {level+1} definition "
                    "is not a dictionary"
                )
            verbosity_def.setdefault("version", logconfig["version"])
            verbosity_def.setdefault("incremental", True)
            if verbosity_def["incremental"] and not _config_is_incremental(
                verbosity_def
            ):
                raise zocalo.ConfigurationError(
                    f"Logging configuration error: verbosity level {level+1} definition "
                    "defines items not allowed in an incremental definition"
                )

        return LoggingIncrementer(logconfig)


def _monkeypatch_graypy() -> None:
    """
    Monkeypatch a helper class into graypy to support log levels in Graylog.
    This translates Python integer level numbers to syslog levels.
    """

    class PythonLevelToSyslogConverter:
        @staticmethod
        def get(level, _):
            if level < 20:
                return 7  # DEBUG
            elif level < 25:
                return 6  # INFO
            elif level < 30:
                return 5  # NOTICE
            elif level < 40:
                return 4  # WARNING
            elif level < 50:
                return 3  # ERROR
            elif level < 60:
                return 2  # CRITICAL
            else:
                return 1  # ALERT

    graypy.handler.SYSLOG_LEVELS = PythonLevelToSyslogConverter()


def _resolve_hostname(host: str) -> str:
    # Attempt to resolve the hostname only once during setup
    # rather than once for each connection, which is the default behaviour.
    # For a UDP handler the default would mean one lookup per logging call.
    try:
        return socket.gethostbyname(host)
    except Exception:
        return host


def GraylogTCPHandler(*, host: str, port: int) -> logging.Handler:
    """
    A handler factory to enable logging to a Graylog server using graypy via TCP.
    """
    _monkeypatch_graypy()
    return graypy.GELFTCPHandler(_resolve_hostname(host), port, level_names=True)


def GraylogUDPHandler(*, host: str, port: int) -> logging.Handler:
    """
    A handler factory to enable logging to a Graylog server using graypy via UDP.
    """
    _monkeypatch_graypy()
    return graypy.GELFUDPHandler(_resolve_hostname(host), port, level_names=True)


class DowngradeFilter(logging.Filter):
    """
    Reduces the level of a log message within certain boundaries.

    The two thresholds are 'reduce_to' and 'only_below'.

    Messages with a level between 'reduce_to' (default: WARNING) and 'only_below'
    (default: CRITICAL) have their log level changed to 'reduce_to'.

    Messages with a level below the 'reduce_to' threshold, or at or above the
    'only_below' threshold are passed through unchanged.
    """

    def __init__(self, reduce_to: str = "WARNING", *, only_below: str = "CRITICAL"):
        super().__init__()

        self._reduce_to_name = reduce_to
        self._reduce_to_value = logging.getLevelName(reduce_to)
        self._only_below_name = only_below
        self._only_below_value = logging.getLevelName(only_below)

        if self._only_below_value <= self._reduce_to_value:
            raise ValueError(
                f"'reduce_to' ({self._reduce_to_name} = {self._reduce_to_value})"
                " must be smaller than"
                f" 'only_below' ({self._only_below_name} = {self._only_below_value})"
            )

    def __repr__(self) -> str:
        return (
            f"<DowngradeFilter reduce_to={self._reduce_to_name}"
            f" only_below={self._only_below_name}>"
        )

    def filter(self, record: logging.LogRecord) -> bool:
        if record.levelno <= self._reduce_to_value:
            return True
        if record.levelno >= self._only_below_value:
            return True
        record.levelno = self._reduce_to_value
        record.levelname = self._reduce_to_name
        return True



from __future__ import annotations

from unittest import mock

import pytest
from workflows.recipe.wrapper import RecipeWrapper
from workflows.transport.offline_transport import OfflineTransport

import zocalo.configuration
from zocalo.service.mailer import Mailer


@pytest.fixture
def zocalo_configuration():
    mock_zc = mock.MagicMock(zocalo.configuration.Configuration)
    mock_zc.smtp = {
        "host": "localhost",
        "port": 4242,
        "from": "zocalo@example.com",
    }
    return mock_zc


@pytest.fixture
def mock_smtp_send_message(mocker):
    mock_smtp = mocker.patch("smtplib.SMTP")
    return mock_smtp.return_value.__enter__.return_value.send_message


def test_mailer_receive_msg(zocalo_configuration, mock_smtp_send_message):
    message = {
        "recipe": {
            "1": {
                "parameters": {
                    "recipients": {
                        "all": "foo@example.com",
                        "select": "bar",
                        "bar": ["bar@example.com"],
                    },
                    "subject": "This is a test email",
                    "content": [
                        "header\n",
                        "{pprint_payload}\n",
                        "footer",
                    ],
                },
            },
        },
        "recipe-pointer": 1,
    }
    header = {
        "message-id": mock.sentinel,
        "subscription": mock.sentinel,
    }

    t = OfflineTransport()
    rw = RecipeWrapper(message=message, transport=t)
    mailer = Mailer()
    mailer._environment = {"config": zocalo_configuration}
    mailer.transport = t
    mailer.start()
    msg = {
        "foo": "bar",
        "ham": "spam",
    }
    mailer.receive_msg(rw, header, msg)
    mock_smtp_send_message.assert_called_once()
    email_msg = mock_smtp_send_message.call_args[0][0]
    assert email_msg["To"] == "bar@example.com, foo@example.com"
    assert email_msg["From"] == "zocalo@example.com"
    assert email_msg["Subject"] == "This is a test email"
    assert (
        email_msg.get_content()
        == """\
header
{'foo': 'bar', 'ham': 'spam'}
footer
"""
    )



from __future__ import annotations

import itertools
from unittest import mock

import workflows.transport
from workflows.transport.common_transport import CommonTransport
from workflows.util import generate_unique_host_id

from zocalo.cli.shutdown import run


def test_shutdown_host(mocker):
    mocked_transport = mocker.MagicMock(CommonTransport)
    mocked_lookup = mocker.patch.object(
        workflows.transport, "lookup", return_value=mocked_transport
    )
    host_prefix = ".".join(generate_unique_host_id().split(".")[:-2])
    hosts = ["uk.ac.diamond.ws123.4567", "ws987.6543"]
    expected_hosts = ["uk.ac.diamond.ws123.4567", f"{host_prefix}.ws987.6543"]
    run(hosts)
    mocked_lookup.assert_called_with("PikaTransport")
    mocked_transport().broadcast.assert_has_calls(
        [
            mock.call("command", {"command": "shutdown", "host": host})
            for host in expected_hosts
        ]
    )


def test_shutdown_services(mocker):
    mocked_transport = mocker.MagicMock(CommonTransport)
    mocked_lookup = mocker.patch.object(
        workflows.transport, "lookup", return_value=mocked_transport
    )
    services = ["Foo", "Bar"]
    run(
        list(itertools.chain.from_iterable([["-s", service] for service in services]))
        + ["-t", "PikaTransport"]
    )
    mocked_lookup.assert_called_with("PikaTransport")
    mocked_transport().broadcast.assert_has_calls(
        [
            mock.call("command", {"command": "shutdown", "service": service})
            for service in services
        ]
    )



from __future__ import annotations

import zocalo.configuration

sample_configuration = """
version: 1

some-constants:
  plugin: storage
  this: that
  one: other

empty-plugin:
  plugin: storage

environments:
  some:
    - some-constants
  empty:
    - empty-plugin
  overwrite:
    plugins:
      - constants
      - sane-constants
  nothing: {}

constants:
  plugin: storage
  order:
    - cream
    - jam
  oxford_comma: undecided

sane-constants:
  plugin: storage
  oxford_comma: yes
  units: metric
"""


def test_assert_plain_configuration_object_does_not_have_storage_attribute():
    zc = zocalo.configuration.from_string(sample_configuration)
    assert zc.storage is None
    zc.activate_environment("nothing")
    assert zc.storage is None


def test_plugin_is_available():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("empty")
    assert zc.storage is not None


def test_simple_plugin_usage():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("some")

    assert zc.storage["this"] == "that"
    assert zc.storage["one"] == "other"


def test_multi_plugin_usage():
    zc = zocalo.configuration.from_string(sample_configuration)
    zc.activate_environment("overwrite")

    assert len(zc.storage["order"]) == 2
    assert zc.storage["units"] == "metric"
    assert zc.storage["oxford_comma"] is True



from __future__ import annotations

import sys
from datetime import datetime
from unittest import mock

import workflows.transport
from workflows.transport.common_transport import CommonTransport

import zocalo.cli.dlq_purge as dlq_purge


def gen_header_activemq(i):
    return {
        "content-length": "2489",
        "expires": "0",
        "destination": "/queue/zocalo.garbage.per_image_analysis",
        "subscription": "1",
        "message-id": f"ID:foo.bar.com-{i}",
        "timestamp": "1633102156582",
    }


def gen_header_rabbitmq(i, use_datetime=True):
    tstamp = 1633962302 + 30 * i
    if use_datetime:
        tstamp = datetime.fromtimestamp(tstamp)
    return {
        "x-death": [{"time": tstamp}],
        "subscription": 1,
        "message-id": f"ID:foo.bar.com-{i}",
    }


def test_dlq_purge_activemq(mocker, tmp_path):
    def mock_subscribe(source, receive_message, acknowledgement):
        for i in range(10):
            header = gen_header_activemq(i)
            message = {
                "foo": f"{i}",
            }
            receive_message(header, message)

    mocked_transport = mocker.MagicMock(CommonTransport)
    mocker.patch.object(workflows.transport, "lookup", return_value=mocked_transport)
    mocked_transport().subscribe = mock_subscribe

    testargs = [
        "prog",
        "--location",
        str(tmp_path),
        "--transport",
        "StompTransport",
        "garbage.per_image_analysis",
    ]
    with mock.patch.object(sys, "argv", testargs):
        dlq_purge.run()

    mocked_transport().ack.assert_has_calls(
        [mock.call(gen_header_activemq(i)) for i in range(10)]
    )

    dlq_dirs = list(tmp_path.iterdir())
    assert len(dlq_dirs) == 1
    assert len(list(dlq_dirs[0].glob("**/*"))) == 10


def test_dlq_purge_rabbitmq(mocker, tmp_path):
    def mock_subscribe(source, receive_message, acknowledgement):
        for i in range(10):
            header = gen_header_rabbitmq(i)
            message = {
                "foo": f"{i}",
            }
            receive_message(header, message)

    mocked_transport = mocker.MagicMock(CommonTransport)
    mocker.patch.object(workflows.transport, "lookup", return_value=mocked_transport)
    mocked_transport().subscribe = mock_subscribe

    testargs = [
        "prog",
        "--location",
        str(tmp_path),
        "--transport",
        "PikaTransport",
        "garbage.per_image_analysis",
    ]
    with mock.patch.object(sys, "argv", testargs):
        dlq_purge.run()

    mocked_transport().ack.assert_has_calls(
        [
            mock.call(
                gen_header_rabbitmq(i, use_datetime=False),
            )
            for i in range(10)
        ]
    )

    dlq_dirs = list(tmp_path.iterdir())
    assert len(dlq_dirs) == 1
    assert len(list(dlq_dirs[0].glob("**/*"))) == 10



from __future__ import annotations

import json
import os
import subprocess
from pathlib import Path
from typing import Dict, Optional

import zocalo


def extended_status_dictionary() -> Dict[str, str]:
    """Returns a dictionary of extra fields that can be appended to log
    messages to provide further relevant contextual information."""
    extended_status = {"zocalo": zocalo.__version__}

    if os.getenv("JOB_ID"):
        # Instance may be running as an SGE cluster job
        extended_status["cluster_JOB_ID"] = os.environ["JOB_ID"]
        if os.getenv("SGE_CELL"):
            extended_status["cluster_SGE_CELL"] = os.environ["SGE_CELL"]
        sge_config = os.getenv("SGE_JOB_SPOOL_DIR")
        if sge_config:
            # This may not be trustworthy, so try to verify against the
            # job configuration. Don't worry if this doesn't work.
            try:
                sge_data = Path(sge_config, "config").read_text()
                for line in sge_data.split("\n"):
                    if line.upper().startswith("SGE_CELL="):
                        extended_status["cluster_SGE_CELL"] = line[9:]
                        break
            except Exception:
                pass

    k8s = get_kubernetes_pod_information()
    if k8s:
        if k8s.get("image"):
            extended_status["container_image"] = k8s["image"]
        extended_status["container_node"] = k8s["node"]

    return extended_status


def get_kubernetes_pod_information() -> Optional[Dict[str, str]]:
    """Detects if instance is running inside a Kubernetes pod, and, if so,
    obtains pod information via the Kubernetes API."""

    serviceaccount = Path("/var/run/secrets/kubernetes.io/serviceaccount")
    cacert = serviceaccount / "ca.crt"
    if not serviceaccount.is_dir() or not cacert.is_file():
        return None
    try:
        namespace = serviceaccount.joinpath("namespace").read_text()
        token = serviceaccount.joinpath("token").read_text()
    except Exception:
        return None
    hostname = os.getenv("HOSTNAME")

    query = subprocess.run(
        (
            "curl",
            "-fs",
            "--cacert",
            cacert,
            "--header",
            f"Authorization: Bearer {token}",
            "-X",
            "GET",
            f"https://kubernetes.default.svc/api/v1/namespaces/{namespace}/pods/{hostname}",
        ),
        capture_output=True,
    )
    if query.returncode or query.stderr:
        return None
    try:
        k8s = json.loads(query.stdout)
    except Exception:
        return None

    k8s_information = {}

    images = {c.get("image") for c in k8s["spec"]["containers"]}
    k8s_information["image"] = ",".join(i for i in images if i)
    k8s_information["node"] = k8s["spec"]["nodeName"]

    return k8s_information



from __future__ import annotations

import base64
import datetime
import enum
import hashlib
import logging
import pathlib
import secrets
import urllib
import urllib.request
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
from pydantic import BaseModel, Field
from workflows.transport import pika_transport

import zocalo.configuration

logger = logging.getLogger("zocalo.util.rabbitmq")


class MessageStats(BaseModel):
    publish: Optional[int] = Field(None, description="Count of messages published.")

    publish_in: Optional[int] = Field(
        None,
        description='Count of messages published "in" to an exchange, i.e. not taking account of routing.',
    )
    publish_out: Optional[int] = Field(
        None,
        description='Count of messages published "out" of an exchange, i.e. taking account of routing.',
    )
    confirm: Optional[int] = Field(None, description="Count of messages confirmed.")
    deliver: Optional[int] = Field(
        None,
        description="Count of messages delivered in acknowledgement mode to consumers.",
    )
    deliver_no_ack: Optional[int] = Field(
        None,
        description="Count of messages delivered in no-acknowledgement mode to consumers.",
    )
    get: Optional[int] = Field(
        None,
        description="Count of messages delivered in acknowledgement mode in response to basic.get.",
    )
    get_no_ack: Optional[int] = Field(
        None,
        description="Count of messages delivered in no-acknowledgement mode in response to basic.get.",
    )
    deliver_get: Optional[int] = Field(
        None, description="Sum of all four of the above."
    )
    redeliver: Optional[int] = Field(
        None,
        description="Count of subset of messages in deliver_get which had the redelivered flag set.",
    )
    drop_unroutable: Optional[int] = Field(
        None, description="Count of messages dropped as unroutable."
    )
    return_unroutable: Optional[int] = Field(
        None, description="Count of messages returned to the publisher as unroutable."
    )


class ConnectionState(enum.Enum):
    starting = "starting"
    tuning = "tuning"
    opening = "opening"
    running = "running"
    flow = "flow"
    blocking = "blocking"
    blocked = "blocked"
    closing = "closing"
    closed = "closed"


class ConnectionInfo(BaseModel):
    """TCP/IP connection statistics."""

    pid: Optional[int] = Field(
        int, description="Id of the Erlang process associated with the connection."
    )
    name: str = Field(..., description="Readable name for the connection.")
    port: int = Field(..., description="Server port.")
    host: str = Field(
        ...,
        description="Server hostname obtained via reverse DNS, or its IP address if reverse DNS failed or was disabled.",
    )
    peer_port: int = Field(..., description="Peer port.")
    peer_host: str = Field(
        ...,
        description="Peer hostname obtained via reverse DNS, or its IP address if reverse DNS failed or was not enabled.",
    )
    ssl: bool = Field(
        ...,
        description="Boolean indicating whether the connection is secured with SSL.",
    )
    ssl_protocol: Optional[str] = Field(
        None, description='SSL protocol (e.g. "tlsv1").'
    )
    ssl_key_exchange: Optional[str] = Field(
        None, description='SSL key exchange algorithm (e.g. "rsa").'
    )
    ssl_cipher: Optional[str] = Field(
        None, description='SSL cipher algorithm (e.g. "aes_256_cbc").'
    )
    ssl_hash: Optional[str] = Field(None, description='SSL hash function (e.g. "sha").')
    peer_cert_subject: Optional[str] = Field(
        None, description="The subject of the peer's SSL certificate, in RFC4514 form."
    )
    peer_cert_issuer: Optional[str] = Field(
        None, description="The issuer of the peer's SSL certificate, in RFC4514 form."
    )
    peer_cert_validity: Optional[str] = Field(
        None, description="The period for which the peer's SSL certificate is valid."
    )
    state: ConnectionState
    channels: int = Field(..., description="Number of channels using the connection.")
    protocol: str = Field(
        ...,
        description="Version of the AMQP protocol in use; currently one of: {0,9,1} {0,8,0}",
    )
    auth_mechanism: str = Field(
        ..., description='SASL authentication mechanism used, such as "PLAIN".'
    )
    user: str = Field(..., description="Username associated with the connection.")
    vhost: str = Field(
        ..., description="Virtual host name with non-ASCII characters escaped as in C."
    )
    timeout: Optional[int] = Field(
        None,
        description="Connection timeout / negotiated heartbeat interval, in seconds.",
    )
    frame_max: int = Field(..., description="Maximum frame size (bytes).")
    channel_max: Optional[int] = Field(
        None, description="Maximum number of channels on this connection."
    )
    # client_properties
    # Informational properties transmitted by the client during connection establishment.
    # recv_oct:
    # Octets received.
    # recv_cnt
    # Packets received.
    # send_oct
    # Octets send.
    # send_cnt
    # Packets sent.
    # send_pend
    # Send queue size.
    connected_at: datetime.datetime = Field(
        ..., description="Date and time this connection was established, as timestamp."
    )


class VHostSpec(BaseModel):
    name: str = Field(
        ...,
        description="The name of the virtual host entry.",
    )
    description: str = ""
    tags: List[str] = Field(default_factory=list)
    tracing: bool = False


class PermissionSpec(BaseModel):
    """Sets user permissions.

    For example, this instructs the RabbitMQ broker to grant the user named
    "janeway" access to the virtual host called "my-vhost", with configure
    permissions on all resources whose names starts with "janeway-", and
    write and read permissions on all resources:

        PermissionSpec(
            vhost="my-vhost",
            user="janeway",
            configure="^janeway-.*",
            write=".*",
            read=".*",
        )
    """

    vhost: str = Field(
        ...,
        description='The name of the virtual host to which to grant the user access, defaulting to "/".',
    )
    user: str = Field(
        ...,
        description="The name of the user to grant access to the specified virtual host.",
    )
    configure: str = Field(
        ...,
        description="A regular expression matching resource names for which the user is granted configure permissions.",
    )
    write: str = Field(
        ...,
        description="A regular expression matching resource names for which the user is granted write permissions.",
    )
    read: str = Field(
        ...,
        description="A regular expression matching resource names for which the user is granted read permissions.",
    )


class NodeType(enum.Enum):
    disc = "disc"
    ram = "ram"


class NodeInfo(BaseModel):
    # applications	List of all Erlang applications running on the node.
    # auth_mechanisms	List of all SASL authentication mechanisms installed on the node.
    # cluster_links	A list of the other nodes in the cluster. For each node, there are details of the TCP connection used to connect to it and statistics on data that has been transferred.
    config_files: Optional[List[pathlib.Path]] = Field(
        None, description="List of config files read by the node."
    )
    # contexts	List of all HTTP listeners on the node.
    db_dir: Optional[pathlib.Path] = Field(
        None, description="Location of the persistent storage used by the node."
    )
    disk_free: int = Field(..., description="Disk free space in bytes.")
    disk_free_alarm: bool = Field(
        ..., description="Whether the disk alarm has gone off."
    )
    disk_free_limit: Optional[int] = Field(
        None, description="Point at which the disk alarm will go off."
    )
    enabled_plugins: Optional[List[str]] = Field(
        None,
        description="List of plugins which are both explicitly enabled and running.",
    )
    # exchange_types	Exchange types available on the node.
    fd_total: int = Field(..., description="File descriptors available.")
    fd_used: int = Field(..., description="Used file descriptors.")
    io_read_avg_time: Optional[int] = Field(
        None,
        ge=0,
        description="Average wall time (milliseconds) for each disk read operation in the last statistics interval.",
    )
    io_read_bytes: Optional[int] = Field(
        None, description="Total number of bytes read from disk by the persister."
    )
    io_read_count: Optional[int] = Field(
        None, description="Total number of read operations by the persister."
    )
    io_reopen_count: Optional[int] = Field(
        None,
        description="Total number of times the persister has needed to recycle file handles between queues. In an ideal world this number will be zero; if the number is large, performance might be improved by increasing the number of file handles available to RabbitMQ.",
    )
    io_seek_avg_time: Optional[int] = Field(
        None,
        description="Average wall time (milliseconds) for each seek operation in the last statistics interval.",
    )
    io_seek_count: Optional[int] = Field(
        None, description="Total number of seek operations by the persister."
    )
    io_sync_avg_time: Optional[int] = Field(
        None,
        description="Average wall time (milliseconds) for each fsync() operation in the last statistics interval.",
    )
    io_sync_count: Optional[int] = Field(
        None, description="Total number of fsync() operations by the persister."
    )
    io_write_avg_time: Optional[int] = Field(
        None,
        description="Average wall time (milliseconds) for each disk write operation in the last statistics interval.",
    )
    io_write_bytes: Optional[int] = Field(
        None, description="Total number of bytes written to disk by the persister."
    )
    io_write_count: Optional[int] = Field(
        None, description="Total number of write operations by the persister."
    )
    log_files: Optional[List[pathlib.Path]] = Field(
        None,
        description='List of log files used by the node. If the node also sends messages to stdout, "<stdout>" is also reported in the list.',
    )
    mem_used: int = Field(..., description="Memory used in bytes.")
    mem_alarm: bool = Field(..., description="Whether the memory alarm has gone off.")
    mem_limit: Optional[int] = Field(
        None, description="Point at which the memory alarm will go off."
    )
    mnesia_disk_tx_count: Optional[int] = Field(
        None,
        description="Number of Mnesia transactions which have been performed that required writes to disk. (e.g. creating a durable queue). Only transactions which originated on this node are included.",
    )
    mnesia_ram_tx_count: Optional[int] = Field(
        None,
        description="Number of Mnesia transactions which have been performed that did not require writes to disk. (e.g. creating a transient queue). Only transactions which originated on this node are included.",
    )
    msg_store_read_count: Optional[int] = Field(
        None,
        description="Number of messages which have been read from the message store.",
    )
    msg_store_write_count: Optional[int] = Field(
        None,
        description="Number of messages which have been written to the message store.",
    )
    name: str = Field(..., description="Node name.")
    net_ticktime: Optional[int] = Field(
        None, description="Current kernel net_ticktime setting for the node."
    )
    os_pid: Optional[int] = Field(
        None,
        description="Process identifier for the Operating System under which this node is running.",
    )
    # partitions	List of network partitions this node is seeing.
    proc_total: int = Field(..., description="Maximum number of Erlang processes.")
    proc_used: int = Field(..., description="Number of Erlang processes in use.")
    processors: Optional[int] = Field(
        None, description="Number of cores detected and usable by Erlang."
    )
    queue_index_journal_write_count: Optional[int] = Field(
        None,
        description="Number of records written to the queue index journal. Each record represents a message being published to a queue, being delivered from a queue, and being acknowledged in a queue.",
    )
    queue_index_read_count: Optional[int] = Field(
        None, description="Number of records read from the queue index."
    )
    queue_index_write_count: Optional[int] = Field(
        None, description="Number of records written to the queue index."
    )
    # rates_mode: 'none', 'basic' or 'detailed'.
    run_queue: float = Field(
        ..., description="Average number of Erlang processes waiting to run."
    )
    running: bool = Field(
        ...,
        description="Boolean for whether this node is up. Obviously if this is false, most other stats will be missing.",
    )
    # sasl_log_file	Location of sasl log file.
    sockets_total: Optional[int] = Field(
        None, description="File descriptors available for use as sockets."
    )
    sockets_used: int = Field(..., description="File descriptors used as sockets.")
    type: Optional[NodeType] = None
    uptime: Optional[int] = Field(
        None, description="Time since the Erlang VM started, in milliseconds."
    )
    # memory	Detailed memory use statistics. Only appears if ?memory=true is appended to the URL.
    # binary	Detailed breakdown of the owners of binary memory. Only appears if ?binary=true is appended to the URL. Note that this can be an expensive query if there are many small binaries in the system.


class DestinationType(enum.Enum):
    QUEUE = "q"
    EXCHANGE = "e"


class BindingSpec(BaseModel):
    source: str = Field(
        ..., description="The name of the source exchange of the binding"
    )
    destination: str = Field(
        ...,
        description="The name of the end point of the binding (either an exchange or a queue)",
    )
    destination_type: DestinationType = Field(
        ..., description="The type of the binding end point"
    )
    vhost: str = Field(
        ..., description="Virtual host name with non-ASCII characters escaped as in C."
    )
    routing_key: str = Field("", description="Routing key attached to binding")
    arguments: Optional[dict] = Field(
        default_factory=dict, description="Binding arguments"
    )


class BindingInfo(BindingSpec):
    properties_key: str = Field(
        "",
        description="Unique identifier composed of the bindings routing key and a hash of its arguments",
    )


class ExchangeType(enum.Enum):
    direct = "direct"
    topic = "topic"
    headers = "headers"
    fanout = "fanout"
    x_delayed_message = "x-delayed-message"


class ExchangeSpec(BaseModel):
    name: str = Field(
        ...,
        description="The name of the exchange with non-ASCII characters escaped as in C.",
    )
    type: ExchangeType = Field(..., description="The exchange type")
    durable: Optional[bool] = Field(
        False, description="Whether or not the exchange survives server restarts."
    )
    auto_delete: Optional[bool] = Field(
        False,
        description="Whether the exchange will be deleted automatically when no longer used.",
    )
    internal: Optional[bool] = Field(
        False,
        description="Whether the exchange is internal, i.e. cannot be directly published to by a client.",
    )
    arguments: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Exchange arguments."
    )
    vhost: str = Field(
        ..., description="Virtual host name with non-ASCII characters escaped as in C."
    )

    class Config:
        use_enum_values = True


class ExchangeInfo(ExchangeSpec):
    policy: Optional[str] = Field(
        None, description="Policy name for applying to the exchange."
    )
    message_stats: Optional[MessageStats] = None
    incoming: Optional[dict] = Field(
        None,
        description="Detailed message stats (see section above) for publishes from channels into this exchange.",
    )
    outgoing: Optional[dict] = Field(
        None,
        description="Detailed message stats for publishes from this exchange into queues.",
    )


class PolicyApplyTo(enum.Enum):
    """Which types of object this policy should apply to."""

    queues = "queues"
    exchanges = "exchanges"
    all = "all"


class PolicySpec(BaseModel):
    """Sets a policy."""

    vhost: str = Field(
        ..., description="Virtual host name with non-ASCII characters escaped as in C."
    )
    name: str = Field(..., description="The name of the policy.")
    pattern: str = Field(
        ...,
        description="The regular expression, which when matches on a given resources causes the policy to apply.",
    )
    definition: Dict[str, Any] = Field(
        ...,
        description="A set of key/value pairs (think a JSON document) that will be injected into the map of optional arguments of the matching queues and exchanges.",
    )
    priority: int = Field(
        0,
        description="The priority of the policy as an integer. Higher numbers indicate greater precedence. The default is 0.",
    )
    apply_to: PolicyApplyTo = Field(
        default=PolicyApplyTo.all,
        alias="apply-to",
        description="Which types of object this policy should apply to.",
    )

    class Config:
        use_enum_values = True
        validate_all = True
        allow_population_by_field_name = True


class QueueState(str, enum.Enum):
    'The state of the queue. Normally "running", but may be "{syncing, message_count}" if the queue is synchronising.'

    running = "running"
    syncing = "syncing"
    message_count = "message_count"


class QueueSpec(BaseModel):
    name: str = Field(
        ...,
        description="The name of the queue with non-ASCII characters escaped as in C.",
    )
    durable: Optional[bool] = Field(
        False, description="Whether or not the queue survives server restarts."
    )
    auto_delete: Optional[bool] = Field(
        False,
        description="Whether the queue will be deleted automatically when no longer used.",
    )
    arguments: Optional[Dict[str, Any]] = Field(
        default_factory=dict, description="Queue arguments."
    )
    vhost: str = Field(
        ..., description="Virtual host name with non-ASCII characters escaped as in C."
    )


class QueueInfo(QueueSpec):
    policy: Optional[str] = Field(
        None, description="Effective policy name for the queue."
    )
    pid: Optional[int] = Field(
        None, description="Erlang process identifier of the queue."
    )
    owner_pid: Optional[int] = Field(
        None,
        description="Id of the Erlang process of the connection which is the exclusive owner of the queue. Empty if the queue is non-exclusive.",
    )
    exclusive: bool = Field(
        ...,
        description="True if queue is exclusive (i.e. has owner_pid), false otherwise.",
    )
    exclusive_consumer_pid: Optional[int] = Field(
        None,
        description="Id of the Erlang process representing the channel of the exclusive consumer subscribed to this queue. Empty if there is no exclusive consumer.",
    )
    exclusive_consumer_tag: Optional[str] = Field(
        None,
        description="Consumer tag of the exclusive consumer subscribed to this queue. Empty if there is no exclusive consumer.",
    )
    messages_ready: Optional[int] = Field(
        None, description="Number of messages ready to be delivered to clients."
    )
    messages_unacknowledged: Optional[int] = Field(
        None,
        description="Number of messages delivered to clients but not yet acknowledged.",
    )
    messages: Optional[int] = Field(
        None, description="Sum of ready and unacknowledged messages (queue depth)."
    )
    messages_ready_ram: Optional[int] = Field(
        None,
        description="Number of messages from messages_ready which are resident in ram.",
    )
    messages_unacknowledged_ram: Optional[int] = Field(
        None,
        description="Number of messages from messages_unacknowledged which are resident in ram.",
    )
    messages_ram: Optional[int] = Field(
        None, description="Total number of messages which are resident in ram."
    )
    messages_persistent: Optional[int] = Field(
        None,
        description="Total number of persistent messages in the queue (will always be 0 for transient queues).",
    )
    message_bytes: Optional[int] = Field(
        None,
        description="Sum of the size of all message bodies in the queue. This does not include the message properties (including headers) or any overhead.",
    )
    message_bytes_ready: Optional[int] = Field(
        None,
        description="Like message_bytes but counting only those messages ready to be delivered to clients.",
    )
    message_bytes_unacknowledged: Optional[int] = Field(
        None,
        description="Like message_bytes but counting only those messages delivered to clients but not yet acknowledged.",
    )
    message_bytes_ram: Optional[int] = Field(
        None,
        description="Like message_bytes but counting only those messages which are currently held in RAM.",
    )
    message_bytes_persistent: Optional[int] = Field(
        None,
        description="Like message_bytes but counting only those messages which are persistent.",
    )
    head_message_timestamp: Optional[datetime.datetime] = Field(
        None,
        description="The timestamp property of the first message in the queue, if present. Timestamps of messages only appear when they are in the paged-in state.",
    )
    disk_reads: Optional[int] = Field(
        None,
        description="Total number of times messages have been read from disk by this queue since it started.",
    )
    disk_writes: Optional[int] = Field(
        None,
        description="Total number of times messages have been written to disk by this queue since it started.",
    )
    consumers: Optional[int] = Field(None, description="Number of consumers.")
    consumer_utilisation: Optional[float] = Field(
        None,
        ge=0,
        le=1,
        description="Fraction of the time (between 0.0 and 1.0) that the queue is able to immediately deliver messages to consumers. This can be less than 1.0 if consumers are limited by network congestion or prefetch count.",
    )
    memory: Optional[int] = Field(
        None,
        description="Bytes of memory allocated by the runtime for the queue, including stack, heap and internal structures.",
    )
    state: Optional[QueueState] = None
    message_stats: Optional[MessageStats] = None
    incoming: Optional[dict] = Field(
        None,
        description="Detailed message stats (see section above) for publishes from exchanges into this queue.",
    )
    deliveries: Optional[dict] = Field(
        None,
        description="Detailed message stats for deliveries from this queue into channels.",
    )
    consumer_details: Optional[List[Any]] = Field(
        None,
        description="List of consumers on this channel, with some details on each.",
    )


class HashingAlgorithm(enum.Enum):
    rabbit_password_hashing_sha256 = "rabbit_password_hashing_sha256"
    rabbit_password_hashing_sha512 = "rabbit_password_hashing_sha512"
    rabbit_password_hashing_md5 = "rabbit_password_hashing_md5"


class UserSpec(BaseModel):
    """
    Either password or password_hash can be set. If neither are set the user will not be
    able to log in with a password, but other mechanisms like client certificates may
    be used. Setting password_hash to "" will ensure the user cannot use a password to
    log in. tags is a list of tags for the user. Currently recognised tags are
    administrator, monitoring, and management.
    """

    name: str = Field(..., description="Username")
    password_hash: str = Field(..., description="Hash of the user password.")
    hashing_algorithm: HashingAlgorithm
    tags: List[str]

    class Config:
        use_enum_values = True


def http_api_request(
    zc: zocalo.configuration.Configuration,
    api_path: str,
) -> urllib.request.Request:
    """
    Return a urllib.request.Request to query the RabbitMQ HTTP API.

    Credentials are obtained via zocalo.configuration.

    Args:
        zc (zocalo.configuration.Configuration): Zocalo configuration object
        api_path (str): The path to be combined with the base_url defined by
            the Zocalo configuration object to give the full path to the API
            endpoint.
    """
    if not zc.rabbitmqapi:
        raise zocalo.ConfigurationError(
            "There are no RabbitMQ API credentials configured in your environment"
        )
    password_mgr = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    password_mgr.add_password(
        realm=None,
        uri=zc.rabbitmqapi["base_url"],
        user=zc.rabbitmqapi["username"],
        passwd=zc.rabbitmqapi["password"],
    )
    handler = urllib.request.HTTPBasicAuthHandler(password_mgr)
    opener = urllib.request.build_opener(handler)
    urllib.request.install_opener(opener)
    return urllib.request.Request(f"{zc.rabbitmqapi['base_url']}{api_path}")


class RabbitMQAPI:
    def __init__(self, url: str, user: str, password: str):
        self._url = url
        self._session = requests.Session()
        self._session.auth = (user, password)

    @classmethod
    def from_zocalo_configuration(cls, zc: zocalo.configuration.Configuration):
        instance = None
        base_url = zc.rabbitmqapi["base_url"]
        for url in base_url.split(","):
            instance = cls(
                url=url,
                user=zc.rabbitmqapi["username"],
                password=zc.rabbitmqapi["password"],
            )
            try:
                instance.get("health/checks/alarms")
                break
            except requests.ConnectionError as e:
                logger.warning(f"Could not connect to {url}: {e}")
                instance = None
        if not instance:
            raise RuntimeError(f"Could not connect to RabbitMQ API: {base_url}")
        return instance

    @property
    def health_checks(self) -> Tuple[Dict[str, Any], Dict[str, str]]:
        # https://rawcdn.githack.com/rabbitmq/rabbitmq-server/v3.9.7/deps/rabbitmq_management/priv/www/api/index.html
        HEALTH_CHECKS = {
            "health/checks/alarms",
            "health/checks/local-alarms",
            "health/checks/certificate-expiration/1/months",
            f"health/checks/port-listener/{pika_transport.PikaTransport.defaults['--rabbit-port']}",
            # f"health/checks/port-listener/1234",
            "health/checks/protocol-listener/amqp",
            "health/checks/virtual-hosts",
            "health/checks/node-is-mirror-sync-critical",
            "health/checks/node-is-quorum-critical",
        }

        success = {}
        failure = {}
        for health_check in HEALTH_CHECKS:
            response = self.get(health_check)
            if response.status_code == requests.codes.ok:
                success[health_check] = response.json()
            else:
                failure[health_check] = response.json()
        return success, failure

    def get(
        self, endpoint: str, params: Dict[str, Any] = None, timeout: float | None = None
    ) -> requests.Response:
        return self._session.get(
            f"{self._url}/{endpoint}", params=params, timeout=timeout
        )

    def put(
        self,
        endpoint: str,
        params: Dict[str, Any] = None,
        json: Dict[str, Any] = None,
        timeout: float | None = None,
    ) -> requests.Response:
        return self._session.put(
            f"{self._url}/{endpoint}", params=params, json=json, timeout=timeout
        )

    def post(
        self,
        endpoint: str,
        data: Optional[Dict[str, Any]] = None,
        json: Optional[Dict[str, Any]] = None,
        timeout: float | None = None,
    ) -> requests.Response:
        return self._session.post(
            f"{self._url}/{endpoint}", data=data, json=json, timeout=timeout
        )

    def delete(
        self, endpoint: str, params: Dict[str, Any] = None, timeout: float | None = None
    ) -> requests.Response:
        return self._session.delete(
            f"{self._url}/{endpoint}", params=params, timeout=timeout
        )

    def bindings(
        self,
        vhost: Optional[str] = None,
        source: Optional[str] = None,
        destination: Optional[str] = None,
        destination_type: Optional[str] = None,
    ) -> List[BindingInfo]:
        endpoint = "bindings"
        if vhost is not None:
            endpoint = f"{endpoint}/{vhost}"
        _check = {source, destination, destination_type}
        if None in _check and len(_check) > 1:
            raise ValueError(
                "Either all of source, destination and destination_type must be specified, or none of them"
            )
        if destination_type is not None:
            endpoint = f"{endpoint}/e/{source}/{destination_type}/{destination}"
        dest_map = {"queue": "q", "exchange": "e"}
        conv = (
            lambda key, value: dest_map[value] if key == "destination_type" else value
        )
        return [
            BindingInfo(**{_r[0]: conv(_r[0], _r[1]) for _r in r.items()})
            for r in self.get(endpoint).json()
        ]

    def binding_declare(self, binding: BindingSpec):
        endpoint = f"bindings/{binding.vhost}/e/{binding.source}/{binding.destination_type.value}/{binding.destination}"
        response = self.post(
            endpoint,
            json=binding.dict(
                exclude_defaults=True,
                exclude={"vhost", "source", "destination", "destination_type"},
            ),
        )
        response.raise_for_status()

    def bindings_delete(
        self,
        vhost: str,
        source: str,
        destination: str,
        destination_type: str,
        properties_key: Optional[str] = None,
    ):
        # If properties_key is not specified then all bindings between the specified
        # source and destination are deleted
        endpoint = f"bindings/{vhost}/e/{source}/{destination_type}/{destination}"
        if properties_key is None:
            dest_map = {"queue": "q", "exchange": "e"}
            conv = (
                lambda key, value: dest_map[value]
                if key == "destination_type"
                else value
            )
            props = [
                BindingInfo(
                    **{_r[0]: conv(_r[0], _r[1]) for _r in r.items()}
                ).properties_key
                for r in self.get(endpoint).json()
            ]
        else:
            props = [properties_key]
        for prop in props:
            response = self.delete(f"{endpoint}/{prop}")
            response.raise_for_status()

    def connections(
        self, name: Optional[str] = None
    ) -> Union[List[ConnectionInfo], ConnectionInfo]:
        endpoint = "connections"
        if name is not None:
            endpoint = f"{endpoint}/{name}/"
            response = self.get(endpoint)
            return ConnectionInfo(**response.json())
        response = self.get(endpoint)
        return [ConnectionInfo(**qi) for qi in response.json()]

    def nodes(self, name: Optional[str] = None) -> Union[List[NodeInfo], NodeInfo]:
        # https://www.rabbitmq.com/monitoring.html#node-metrics
        endpoint = "nodes"
        if name is not None:
            endpoint = f"{endpoint}/{name}"
            response = self.get(endpoint)
            return NodeInfo(**response.json())
        response = self.get(endpoint)
        return [NodeInfo(**qi) for qi in response.json()]

    def exchanges(
        self, vhost: Optional[str] = None, name: Optional[str] = None
    ) -> Union[List[ExchangeInfo], ExchangeInfo]:
        endpoint = "exchanges"
        if vhost is not None and name is not None:
            endpoint = f"{endpoint}/{vhost}/{name}/"
            response = self.get(endpoint)
            return ExchangeInfo(**response.json())
        elif vhost is not None:
            endpoint = f"{endpoint}/{vhost}/"
        elif name is not None:
            raise ValueError("name can not be set without vhost")
        response = self.get(endpoint)
        return [ExchangeInfo(**qi) for qi in response.json()]

    def exchange_declare(self, exchange: ExchangeSpec):
        endpoint = f"exchanges/{exchange.vhost}/{exchange.name}/"
        response = self.put(
            endpoint,
            json=exchange.dict(exclude_defaults=True, exclude={"name", "vhost"}),
        )
        response.raise_for_status()

    def exchange_delete(self, vhost: str, name: str, if_unused: bool = False):
        endpoint = f"exchanges/{vhost}/{name}"
        response = self.delete(endpoint, params={"if-unused": if_unused})
        response.raise_for_status()

    def policies(self, vhost: Optional[str] = None) -> List[PolicySpec]:
        endpoint = "policies"
        if vhost is not None:
            endpoint = f"{endpoint}/{vhost}/"
        response = self.get(endpoint)
        return [PolicySpec(**p) for p in response.json()]

    def policy(self, vhost: str, name: str) -> PolicySpec:
        endpoint = f"policies/{vhost}/{name}/"
        response = self.get(endpoint)
        return PolicySpec(**response.json())

    def set_policy(self, policy: PolicySpec):
        endpoint = f"policies/{policy.vhost}/{policy.name}/"
        response = self.put(
            endpoint,
            json=policy.dict(
                exclude_defaults=True, exclude={"name", "vhost"}, by_alias=True
            ),
        )
        response.raise_for_status()

    def clear_policy(self, vhost: str, name: str):
        endpoint = f"policies/{vhost}/{name}/"
        response = self.delete(endpoint)
        response.raise_for_status()

    def queues(
        self, vhost: Optional[str] = None, name: Optional[str] = None
    ) -> Union[List[QueueInfo], QueueInfo]:
        endpoint = "queues"
        if vhost is not None and name is not None:
            endpoint = f"{endpoint}/{vhost}/{name}"
            response = self.get(endpoint)
            return QueueInfo(**response.json())
        elif vhost is not None:
            endpoint = f"{endpoint}/{vhost}"
        elif name is not None:
            raise ValueError("name can not be set without vhost")
        response = self.get(endpoint)
        return [QueueInfo(**qi) for qi in response.json()]

    def queue_declare(self, queue: QueueSpec):
        endpoint = f"queues/{queue.vhost}/{queue.name}"
        response = self.put(
            endpoint, json=queue.dict(exclude_defaults=True, exclude={"name", "vhost"})
        )
        response.raise_for_status()

    def queue_delete(
        self, vhost: str, name: str, if_unused: bool = False, if_empty: bool = False
    ):
        endpoint = f"queues/{vhost}/{name}"
        response = self.delete(
            endpoint, params={"if-unused": if_unused, "if-empty": if_empty}
        )
        response.raise_for_status()

    def users(self) -> List[UserSpec]:
        endpoint = "users"
        response = self.get(endpoint)
        return [UserSpec(**user) for user in response.json()]

    def user(self, name: str) -> UserSpec:
        endpoint = f"users/{name}/"
        response = self.get(endpoint).json()
        return UserSpec(**response)

    def permissions(
        self, vhost: Optional[str] = None, user: Optional[str] = None
    ) -> List[PermissionSpec] | PermissionSpec:
        endpoint = "permissions"
        if vhost is not None and user is not None:
            endpoint = f"{endpoint}/{vhost}/{user}/"
            response = self.get(endpoint)
            return PermissionSpec(**response.json())
        elif vhost is not None or user is not None:
            raise ValueError(
                "Either both or neither of vhost and user must be set to None"
            )
        response = self.get(endpoint)
        return [PermissionSpec(**ps) for ps in response.json()]

    def set_permissions(self, permission: PermissionSpec):
        endpoint = f"permissions/{permission.vhost}/{permission.user}/"
        submission = permission.dict(exclude_defaults=True, exclude={"vhost", "user"})
        response = self.put(endpoint, json=submission)
        response.raise_for_status()

    def clear_permissions(self, vhost: str, user: str):
        endpoint = f"permissions/{vhost}/{user}/"
        response = self.delete(endpoint)
        response.raise_for_status()

    def user_put(self, user: UserSpec):
        endpoint = f"users/{user.name}/"
        submission = user.dict(exclude_defaults=True, exclude={"name"})
        submission["tags"] = ",".join(submission["tags"])
        response = self.put(endpoint, json=submission)
        response.raise_for_status()

    def user_delete(self, name: str):
        endpoint = f"users/{name}/"
        response = self.delete(endpoint)
        response.raise_for_status()

    def vhosts(self) -> List[VHostSpec]:
        endpoint = "vhosts"
        response = self.get(endpoint)
        return [VHostSpec(**user) for user in response.json()]

    def vhost(self, name: str) -> VHostSpec:
        endpoint = f"vhosts/{name}/"
        response = self.get(endpoint).json()
        return VHostSpec(**response)

    def add_vhost(self, vhost: VHostSpec):
        endpoint = f"vhosts/{vhost.name}/"
        response = self.put(
            endpoint, json=vhost.dict(exclude_defaults=True, exclude={"name", "vhost"})
        )
        response.raise_for_status()

    def delete_vhost(self, name: str):
        endpoint = f"vhosts/{name}/"
        response = self.delete(endpoint)
        response.raise_for_status()


def hash_password(passwd: str, salt: Optional[str] = None) -> str:
    if salt:
        # extract salt from an existing password hash
        salt_bytes = base64.b64decode(salt)[:4]
    else:
        salt_bytes = secrets.token_bytes(4)
    utf8 = passwd.encode("utf-8")
    temp1 = salt_bytes + utf8
    temp2 = hashlib.sha256(temp1).digest()  # lgtm
    salted_hash = salt_bytes + temp2
    pass_hash = base64.b64encode(salted_hash)
    return pass_hash.decode()



from __future__ import annotations

import zocalo.util


def test_can_extract_kubernetes_information():
    kubernetes = zocalo.util.get_kubernetes_pod_information()

    if kubernetes is None:
        return
    assert kubernetes["node"]


def test_extended_status():
    status = zocalo.util.extended_status_dictionary()

    assert status["zocalo"] == zocalo.__version__



from __future__ import annotations

from marshmallow import fields

from zocalo.configuration import PluginSchema


class RabbitAPI:
    class Schema(PluginSchema):
        base_url = fields.Str(required=True)
        username = fields.Str(required=True)
        password = fields.Str(required=True)
        vhost = fields.Str()

    @staticmethod
    def activate(configuration):
        configuration.setdefault("vhost", "/")
        return configuration



from __future__ import annotations

from unittest import mock

import workflows.transport
from workflows.transport.common_transport import CommonTransport

from zocalo.cli.queue_drain import run


def test_queue_drain(mocker, capsys):
    filtered_header = {
        "priority": "4",
        "workflows-recipe": "True",
        "persistent": "true",
    }

    def mock_subscribe(source, receive_message, acknowledgement):
        for i in range(10):
            header = {
                **filtered_header,
                "content-length": "2489",
                "expires": "0",
                "destination": "/queue/zocalo.garbage.per_image_analysis",
                "subscription": "1",
                "message-id": f"ID:foo.bar.com-{i}",
                "timestamp": "1633102156582",
            }
            message = {
                "foo": f"{i}",
            }
            receive_message(header, message)

    mocked_transport = mocker.MagicMock(CommonTransport)
    mocker.patch.object(workflows.transport, "lookup", return_value=mocked_transport)
    mocked_transport().subscribe = mock_subscribe

    run(["source", "destination", "--wait", "0.1", "--stop", "1.5"])

    captured = capsys.readouterr()
    assert "10 message(s) drained, no message seen for" in captured.out

    mocked_transport().send.assert_has_calls(
        [
            mock.call(
                "destination",
                {"foo": f"{i}"},
                headers=filtered_header,
                transaction=mock.ANY,
            )
            for i in range(10)
        ]
    )



#
# zocalo.go
#   Process a datacollection
#

from __future__ import annotations

import argparse
import getpass
import json
import pathlib
import socket
import sys
import uuid
from pprint import pprint

import workflows.recipe
import workflows.transport

import zocalo.configuration.argparse

# Example: zocalo.go -r example-xia2 527189


def run():
    zc = zocalo.configuration.from_file()
    zc.activate()

    parser = argparse.ArgumentParser(
        usage="zocalo.go [options] dcid",
        description="Triggers processing of a standard "
        "recipe, of an arbitrary recipe from a local file, or of an entry in "
        "the ISPyB processing table.",
    )

    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "-r",
        "--recipe",
        dest="recipe",
        metavar="RCP",
        action="append",
        default=[],
        help="Name of a recipe to run. Can be used multiple times. "
        "Recipe names correspond to filenames (excluding .json) in /dls_sw/apps/zocalo/live/recipes",
    )
    parser.add_argument(
        "-a",
        "--autoprocscalingid",
        dest="autoprocscalingid",
        metavar="APSID",
        action="store",
        type=str,
        default=None,
        help="An auto processing scaling ID for downstream processing recipes.",
    )
    parser.add_argument(
        "-f",
        "--file",
        dest="recipefile",
        metavar="FILE",
        action="store",
        type=str,
        default="",
        help="Run recipe contained in this file.",
    )
    parser.add_argument(
        "-n",
        "--no-dcid",
        dest="nodcid",
        action="store_true",
        default=False,
        help="Trigger recipe without specifying a data collection ID",
    )
    parser.add_argument(
        "--drop",
        dest="dropfile",
        action="store_true",
        default=False,
        help=argparse.SUPPRESS,
    )  # Write directly to file, do not attempt to send as message
    parser.add_argument(
        "-p",
        "--reprocessing",
        dest="reprocess",
        action="store_true",
        default=False,
        help="Means a reprocessing ID is given rather than a data collection ID",
    )
    parser.add_argument(
        "-s",
        "--set",
        dest="parameters",
        action="append",
        default=[],
        metavar="KEY=VALUE",
        help="Set an additional variable for recipe evaluation",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        default=False,
        help="Show raw message before sending",
    )
    parser.add_argument(
        "--dry-run",
        dest="dryrun",
        action="store_true",
        default=False,
        help="Verify that everything is in place that the message could be sent, but don't actually send the message",
    )
    parser.add_argument(
        "dcid", nargs="*", help="Data collection ID of required processing"
    )

    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)
    args = parser.parse_args()

    if zc.storage and zc.storage.get("zocalo.go.fallback_location"):
        dropfile_fallback = pathlib.Path(zc.storage["zocalo.go.fallback_location"])
    else:
        dropfile_fallback = False

    def write_message_to_dropfile(message, headers):
        message_serialized = (
            json.dumps({"headers": headers, "message": message}, indent=2) + "\n"
        )

        fallback = dropfile_fallback / str(uuid.uuid4())
        if args.dryrun:
            print("Not storing message in %s (running with --dry-run)" % fallback)
            return
        fallback.write_text(message_serialized)
        print("Message successfully stored in %s" % fallback)

    def send_or_defer(message):
        headers = {
            "zocalo.go.user": getpass.getuser(),
            "zocalo.go.host": socket.gethostname(),
        }
        if args.verbose:
            pprint(message)
        if dropfile_fallback and args.dropfile:
            return write_message_to_dropfile(message, headers)
        try:
            transport = workflows.transport.lookup(args.transport)()
            if args.dryrun:
                print("Not sending message (running with --dry-run)")
                return
            transport.connect()
            transport.send("processing_recipe", message, headers=headers)
            transport.disconnect()
        except (
            KeyboardInterrupt,
            SyntaxError,
            AssertionError,
            AttributeError,
            ImportError,
            TypeError,
            ValueError,
        ):
            raise
        except Exception:
            if not dropfile_fallback:
                raise
            print("\n\n")
            import traceback

            traceback.print_exc()
            print("\n\nAttempting to store message in fallback location")
            write_message_to_dropfile(message, headers)

    message = {"recipes": args.recipe, "parameters": {}}
    for kv in args.parameters:
        if "=" not in kv:
            sys.exit(f"Invalid variable specification '{kv}'")
        key, value = kv.split("=", 1)
        message["parameters"][key] = value

    if (
        not args.recipe
        and not args.recipefile
        and not args.nodcid
        and not args.reprocess
    ):
        sys.exit("No recipes specified.")

    if args.recipefile:
        with open(args.recipefile) as fh:
            custom_recipe = workflows.recipe.Recipe(json.load(fh))
        custom_recipe.validate()
        message["custom_recipe"] = custom_recipe.recipe

    if args.nodcid:
        if args.recipe:
            print("Running recipes", args.recipe)
        if args.recipefile:
            print("Running recipe from file", args.recipefile)
        print("without specified data collection.")
        send_or_defer(message)
        print("\nSubmitted.")
        sys.exit(0)

    if len(args.dcid) > 1:
        sys.exit("Only a single data collection ID can be specified.")
    if not args.dcid:
        sys.exit("You must either specify a data collection ID or option --no-dcid.")

    dcid = int(args.dcid[0])
    assert dcid > 0, "Invalid data collection ID given."

    if args.reprocess:
        # Given ID is a reprocessing ID. Nothing else needs to be specified.
        if args.recipe:
            print("Running recipes", args.recipe)
        message["parameters"]["ispyb_process"] = dcid
        send_or_defer(message)
        print("\nReprocessing task submitted for ID %d." % dcid)
        sys.exit(0)

    if message["recipes"]:
        print("Running recipes", message["recipes"])

    if args.recipefile:
        print("Running recipe from file", args.recipefile)

    if not message["recipes"] and not message.get("custom_recipe"):
        sys.exit("No recipes specified.")
    print("for data collection", dcid)
    message["parameters"]["ispyb_dcid"] = dcid

    if args.autoprocscalingid:
        apsid = int(args.autoprocscalingid)
        assert apsid > 0, "Invalid auto processing scaling ID given."
        message["parameters"]["ispyb_autoprocscalingid"] = apsid

    send_or_defer(message)
    print("\nSubmitted.")



from __future__ import annotations


class Storage:
    @staticmethod
    def activate(configuration, config_object):
        storage = config_object.storage or {}
        storage.update(configuration)
        return storage



from __future__ import annotations

import zocalo.util.symlink


def test_create_parent_symlink(tmp_path):
    d = tmp_path / "123456-uuid-7890" / "tool"
    d.mkdir(parents=True)
    assert zocalo.util.symlink.create_parent_symlink(d, "some-name") is True
    assert tmp_path.joinpath("some-name").is_symlink()
    assert tmp_path.joinpath("some-name").resolve() == d.resolve()
    assert (
        zocalo.util.symlink.create_parent_symlink(
            d, "some-name", overwrite_symlink=True
        )
        is True
    )
    assert tmp_path.joinpath("some-name").resolve() == d.resolve()

    d = tmp_path / "123456-uuid-7890" / "tool-other"
    d.mkdir()
    assert zocalo.util.symlink.create_parent_symlink(d, "some-name") is False
    assert tmp_path.joinpath("some-name").resolve() != d.resolve()
    assert (
        zocalo.util.symlink.create_parent_symlink(
            d, "some-name", overwrite_symlink=True
        )
        is True
    )
    assert tmp_path.joinpath("some-name").resolve() == d.resolve()


def test_create_parent_symlink_levels(tmp_path):
    d = tmp_path / "123456-uuid-7890" / "tool"
    d.mkdir(parents=True)
    assert zocalo.util.symlink.create_parent_symlink(d, "some-name", levels=1) is True
    assert (tmp_path / "123456-uuid-7890" / "some-name").is_symlink()
    assert (tmp_path / "123456-uuid-7890" / "some-name").resolve() == d.resolve()


def test_create_parent_symlink_does_not_overwrite_files(tmp_path):
    d = tmp_path / "123456-uuid-7890" / "tool"
    d.mkdir(parents=True)
    tmp_path.joinpath("some-name").touch()
    assert zocalo.util.symlink.create_parent_symlink(d, "some-name") is False
    assert (
        zocalo.util.symlink.create_parent_symlink(
            d, "some-name", overwrite_symlink=True
        )
        is False
    )
    assert not tmp_path.joinpath("some-name").is_symlink()



from __future__ import annotations

from marshmallow import fields

from zocalo.configuration import PluginSchema


class SMTP:
    class Schema(PluginSchema):
        host = fields.Str(required=True)
        port = fields.Int(required=True)
        from_ = fields.Email(data_key="from")

    @staticmethod
    def activate(configuration):
        return configuration



from __future__ import annotations

from unittest import mock

from workflows.recipe.wrapper import RecipeWrapper
from workflows.transport.offline_transport import OfflineTransport

from zocalo.service.jsonlines import JSONLines


def test_jsonlines(tmp_path):
    output_file = tmp_path / "foo.json"
    message = {
        "recipe": {
            "1": {
                "parameters": {
                    "output_filename": str(output_file),
                    "include": ["ham", "spam", "bar"],
                    "exclude": ["bar"],
                },
            },
        },
        "recipe-pointer": 1,
    }
    header = {
        "message-id": mock.sentinel,
        "subscription": mock.sentinel,
    }

    t = OfflineTransport()
    rw = RecipeWrapper(message=message, transport=t)
    jsonlines = JSONLines()
    jsonlines.transport = t
    jsonlines.start()
    messages = [
        {"foo": "bar", "ham": 1, "spam": 2, "bar": "foo"},
        {"foo": "bar", "ham": 2, "spam": 3, "bar": "foo"},
    ]
    for msg in messages:
        jsonlines.receive_msg(rw, header, msg)
    # This would normally be called internally by the service
    jsonlines.process_messages()

    content = output_file.read_text()
    assert (
        content
        == """\
{"ham": 1, "spam": 2}
{"ham": 2, "spam": 3}
"""
    )



from __future__ import annotations

import logging
import threading
from typing import Any, Callable

import workflows.util

import zocalo


class BaseWrapper:

    _logger_name = "zocalo.wrapper"  # The logger can be accessed via self.log

    def __init__(self, *args, **kwargs):
        self._environment = kwargs.get("environment", {})
        self.__log_extra = {}
        logger = logging.getLogger(self._logger_name)
        self.log = logging.LoggerAdapter(logger, extra=self.__log_extra)

    def set_recipe_wrapper(self, recwrap):
        self.recwrap = recwrap
        self.__log_extra["recipe_ID"] = recwrap.environment["ID"]

    def prepare(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("starting", payload)

    def update(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("updates", payload)

    def done(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("completed", payload)

    def success(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("success", payload)

    def failure(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("failure", str(payload))

    def run(self):
        raise NotImplementedError()

    def record_result_individual_file(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("result-individual-file", payload)

    def record_result_all_files(self, payload=""):
        if getattr(self, "recwrap", None):
            self.recwrap.send_to("result-all-files", payload)

    @property
    def config(self):
        return self._environment.get("config")


class DummyWrapper(BaseWrapper):

    _logger_name = "zocalo.wrapper.DummyWrapper"

    def run(self):
        self.log.info("This is a dummy wrapper that simply waits for twenty seconds.")
        import time

        time.sleep(10)
        self.status_thread.taskname += " (still running)"
        time.sleep(10)
        return True


class StatusNotifications(threading.Thread):
    def __init__(self, send_function: Callable[[dict], None], taskname: str):
        """Construct and start a StatusNotifications thread object.

        Once started, this will repeatedly re-broadcast the cached
        current status, every few seconds.

        Args:
            send_function: The transport function that broadcasts the actual message.
            taskname: The name of this task, which will be appended to the message.
        """
        super().__init__(name="zocalo status notification")
        self.daemon = True
        self._send_status = send_function
        self._lock = threading.Condition(threading.Lock())
        self._status_dict = {
            "host": workflows.util.generate_unique_host_id(),
            "task": taskname,
            "workflows": workflows.version(),
            "zocalo": zocalo.__version__,
        }
        self.set_status(workflows.services.common_service.Status.STARTING)
        self._keep_running = True
        self.start()

    def set_static_status_field(self, field: str, value: Any) -> None:
        """
        Add an additional static field to status notifications.
        """
        with self._lock:
            self._status_dict[field] = value

    def set_status(self, status: workflows.services.common_service.Status) -> None:
        with self._lock:
            self._status_dict["status"], self._status_dict["statustext"] = (
                status.intval,
                status.description,
            )
            self._lock.notify()

    @property
    def taskname(self) -> str:
        """Return the name displayed on service monitors for this task."""
        return self._status_dict["task"]

    @taskname.setter
    def taskname(self, value: str) -> None:
        """Set/update the name displayed on service monitors for this task."""
        with self._lock:
            self._status_dict["task"] = value
            self._lock.notify()

    def shutdown(self) -> None:
        """Stop the status notification thread."""
        self._keep_running = False

    def send_status(self, dictionary: dict) -> None:
        try:
            self._send_status(dictionary)
        except workflows.Disconnected:
            pass

    def run(self) -> None:
        """Status notification thread main loop."""
        with self._lock:
            self.send_status(self._status_dict)
            while self._keep_running:
                self._lock.wait(3)
                self.send_status(self._status_dict)



#
# zocalo.queue_drain
#   Drain one queue into another in a controlled manner
#


from __future__ import annotations

import argparse
import queue
import sys
import time
from datetime import datetime

import workflows.recipe.wrapper
import workflows.transport

import zocalo.configuration


def show_cluster_info(step):
    try:
        print("Beamline " + step["parameters"]["cluster_project"].upper())
    except Exception:
        pass
    try:
        print("Working directory " + step["parameters"]["workingdir"])
    except Exception:
        pass


show_additional_info = {"cluster.submission": show_cluster_info}


def run(args=None):
    # Load configuration
    zc = zocalo.configuration.from_file()
    zc.activate()

    parser = argparse.ArgumentParser(
        usage="zocalo.queue_drain [options] source destination"
    )
    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument("SOURCE", type=str, help="Source queue name")
    parser.add_argument(
        "DEST",
        type=str,
        help="Destination queue name;"
        " use '.' to automatically determine destination for recipe messages",
    )
    parser.add_argument(
        "--wait",
        action="store",
        dest="wait",
        type=float,
        default=5,
        help="Wait this many seconds between deliveries",
    )
    parser.add_argument(
        "--stop",
        action="store",
        dest="stop",
        type=float,
        default=60,
        help="Stop if no message seen for this many seconds (0 = forever)",
    )
    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)
    args = parser.parse_args(args)

    transport = workflows.transport.lookup(args.transport)()
    transport.connect()

    messages = queue.Queue()

    def receive_message(header, message):
        messages.put((header, message))

    print(f"Reading messages from {args.SOURCE}")
    subscription_id = transport.subscribe(
        args.SOURCE, receive_message, acknowledgement=True
    )

    if args.DEST == ".":
        print("Writing messages to automatically determined destinations")
    else:
        print(f"Writing messages to {args.DEST}")

    message_count = 0
    header_filter = frozenset(
        {
            "content-length",
            "destination",
            "expires",
            "message-id",
            "original-destination",
            "originalExpiration",
            "subscription",
            "timestamp",
            "redelivered",
        }
    )
    drain_start = time.time()
    idle_time = 0
    try:
        while True:
            try:
                header, message = messages.get(True, 0.1)
            except queue.Empty:
                idle_time = idle_time + 0.1
                if args.stop and idle_time > args.stop:
                    break
                continue
            idle_time = 0
            print()
            try:
                print(
                    "Message date: {:%Y-%m-%d %H:%M:%S}".format(
                        datetime.fromtimestamp(int(header["timestamp"]) / 1000)
                    )
                )
            except Exception:
                pass
            target_queue = args.DEST

            try:
                print(f"Recipe ID:    {message['environment']['ID']}")
                r = workflows.recipe.wrapper.RecipeWrapper(message=message)
                if target_queue == ".":
                    target_queue = r.recipe_step["queue"]
                    print(f"Target Queue: {target_queue}")
                additional_info_function = show_additional_info.get(target_queue)
                if additional_info_function:
                    additional_info_function(r.recipe_step)
            except Exception:
                pass

            if target_queue == ".":
                exit("Could not determine target queue for message")

            new_headers = {
                key: header[key] for key in header if key not in header_filter
            }
            txn = transport.transaction_begin(subscription_id=subscription_id)
            transport.send(target_queue, message, headers=new_headers, transaction=txn)
            transport.ack(header, transaction=txn)
            transport.transaction_commit(txn)
            message_count = message_count + 1
            print(
                "%4d message(s) drained in %.1f seconds"
                % (message_count, time.time() - drain_start)
            )
            time.sleep(args.wait)
    except KeyboardInterrupt:
        sys.exit(
            "\nCancelling, %d message(s) drained, %d message(s) unprocessed in memory"
            % (message_count, messages.qsize())
        )
    print(
        "%d message(s) drained, no message seen for %.1f seconds"
        % (message_count, idle_time)
    )
    transport.disconnect()



from __future__ import annotations

import argparse

import workflows.transport

import zocalo.configuration
from zocalo.util.jmxstats import JMXAPI
from zocalo.util.rabbitmq import RabbitMQAPI

#
# zocalo.dlq_check
#   Check number of messages in dead letter queues
#


def check_dlq(zc: zocalo.configuration.Configuration, namespace: str = None) -> dict:
    """Monitor ActiveMQ queue activity."""
    jmx = JMXAPI(zc)
    if namespace:
        namespace = namespace + "."
    else:
        namespace = ""
    result = jmx.org.apache.activemq(
        type="Broker",
        brokerName="localhost",
        destinationType="Queue",
        destinationName=f"DLQ.{namespace}*",
        attribute="QueueSize",
    )
    if result["status"] == 404:
        return {}
    assert result["status"] == 200, result

    def extract_queue_name(namestring):
        namestringdict = {
            component.split("=")[0]: component.split("=", 1)[1]
            for component in namestring.split(",")
            if "=" in component
        }
        return namestringdict.get("destinationName")

    queuedata = {
        extract_queue_name(name): data["QueueSize"]
        for name, data in result["value"].items()
    }
    return queuedata


def check_dlq_rabbitmq(
    zc: zocalo.configuration.Configuration, namespace: str = None
) -> dict:
    rmq = RabbitMQAPI.from_zocalo_configuration(zc)
    return {
        q.name: q.messages
        for q in rmq.queues()
        if q.name.startswith("dlq.")
        and (namespace is None or q.vhost == namespace)
        and q.messages
    }


def run() -> None:
    zc = zocalo.configuration.from_file()
    zc.activate()
    parser = argparse.ArgumentParser(usage="zocalo.dlq_check [options]")
    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "-n",
        "--namespace",
        type=str,
        dest="namespace",
        default="",
        help="Restrict check to this namespace",
    )
    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)
    args = parser.parse_args()

    if args.transport == "StompTransport":
        dlqs = check_dlq(zc, namespace=args.namespace)
        for queue, count in dlqs.items():
            print(f"DLQ for {queue.replace('DLQ.', '')} contains {count} entries")
    elif args.transport == "PikaTransport":
        dlqs = check_dlq_rabbitmq(zc, namespace=args.namespace or "zocalo")
        for queue, count in dlqs.items():
            print(f"DLQ for {queue.replace('dlq.', '')} contains {count} entries")
    else:
        exit(f"Transport {args.transport} not recognised")
    total = sum(dlqs.values())
    if total:
        exit(f"Total of {total} DLQ messages found")
    else:
        print("No DLQ messages found")



from __future__ import annotations

import argparse
import configparser
import functools
import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import requests
import yaml
from pydantic import BaseModel

import zocalo.configuration
from zocalo.util.rabbitmq import (
    BindingSpec,
    ExchangeSpec,
    PermissionSpec,
    PolicySpec,
    QueueSpec,
)
from zocalo.util.rabbitmq import RabbitMQAPI as _RabbitMQAPI
from zocalo.util.rabbitmq import UserSpec, VHostSpec, hash_password

logger = logging.getLogger("zocalo.cli.configure_rabbitmq")


class RabbitMQAPI(_RabbitMQAPI):
    @functools.singledispatchmethod  # type: ignore
    def create_component(self, component: BaseModel):
        raise NotImplementedError(f"Component {component} not recognised")

    @functools.singledispatchmethod  # type: ignore
    def delete_component(self, component: BaseModel):
        raise NotImplementedError(f"Component {component} not recognised")

    @create_component.register  # type: ignore
    def _(self, binding: BindingSpec):
        self.binding_declare(binding)

    @delete_component.register  # type: ignore
    def _(self, binding: BindingSpec):
        self.bindings_delete(
            vhost=binding.vhost,
            source=binding.source,
            destination=binding.destination,
            destination_type=binding.destination_type.value,
        )

    @create_component.register  # type: ignore
    def _(self, exchange: ExchangeSpec):
        self.exchange_declare(exchange)

    @delete_component.register  # type: ignore
    def _(self, exchange: ExchangeSpec, **kwargs):
        self.exchange_delete(vhost=exchange.vhost, name=exchange.name, **kwargs)

    @create_component.register  # type: ignore
    def _(self, vhost: VHostSpec):
        self.add_vhost(vhost)

    @delete_component.register  # type: ignore
    def _(self, vhost: VHostSpec):
        self.delete_vhost(name=vhost.name)

    @create_component.register  # type: ignore
    def _(self, permissions: PermissionSpec):
        self.set_permissions(permissions)

    @delete_component.register  # type: ignore
    def _(self, permissions: PermissionSpec):
        self.clear_permissions(vhost=permissions.vhost, user=permissions.user)


@functools.singledispatch
def _info_to_spec(incoming, infos: list):
    cls = type(incoming)
    return [cls(**i.dict()) for i in infos]


@functools.singledispatch
def _skip(comp) -> bool:
    return False


@_skip.register  # type: ignore
def _(comp: ExchangeSpec) -> bool:
    if comp.name == "" or "amq." in comp.name:
        return True
    return False


@_skip.register  # type: ignore
def _(comp: BindingSpec) -> bool:
    if comp.source == "" or "amq." in comp.source:
        return True
    return False


def update_config(
    api: RabbitMQAPI, incoming: List[BaseModel], current: List[BaseModel]
):
    cls = type(incoming[0])
    current = _info_to_spec(incoming[0], current)
    for cc in current:
        if cc in incoming:
            if hasattr(cc, "name"):
                logger.debug(f"{cls.__name__} {cc.name} already exists")
            elif hasattr(cc, "source"):
                logger.debug(
                    f"{cls.__name__} {cc.source or 'default'}->{cc.destination} already exists"
                )
        else:
            if _skip(cc):
                continue
            logger.info(f"deleting {cls.__name__} {cc}")
            api.delete_component(cc)
    for ic in incoming:
        if ic not in current:
            logger.info(f"creating {cls.__name__} {ic}")
            api.create_component(ic)


def get_vhost_specs(vhosts: dict) -> List[VHostSpec]:
    vhost_specs = []
    for vhost in vhosts:
        vhost_specs.append(
            VHostSpec(
                name=vhost["name"],
                description=vhost.get("description", ""),
                tags=vhost.get("tags", []),
                tracing=vhost.get("tracing", False),
            )
        )
    return vhost_specs


def get_permission_specs(permissions: dict) -> List[PermissionSpec]:
    permission_specs = []
    for permission in permissions:
        permission_specs.append(PermissionSpec(**permission))
    return permission_specs


def get_binding_specs(bindings: dict) -> List[BindingSpec]:
    binding_specs = []
    for binding in bindings:
        binding_specs.append(
            BindingSpec(
                source=binding["source"],
                destination=binding["destination"],
                destination_type=binding.get("destination_type", "q"),
                routing_key=binding.get("routing_key", binding["destination"]),
                vhost=binding["vhost"],
                arguments=binding.get("arguments", {}),
            )
        )
    return binding_specs


def get_binding_specs_for_group(group: dict) -> List[BindingSpec]:
    sources = group.get("bindings", [""])
    vhost = group.get("vhost", "/")
    return [
        BindingSpec(
            vhost=vhost,
            source=source,
            destination=name,
            destination_type="q",
            routing_key=name,
            arguments={},
            properties_key=name,
        )
        for source in sources
        for name in group["names"]
    ]


def get_queue_specs(group: dict) -> List[QueueSpec]:
    queue_settings = group.get("settings", {}).get("queues", {})
    qtype = queue_settings.get("type", "classic")
    dlq_pattern = queue_settings.get("dead-letter-routing-key-pattern")
    dlq_exchange = queue_settings.get("dead-letter-exchange", "")
    dlq_create = queue_settings.get("dead-letter-queue-create", True)
    vhost = group.get("vhost", "/")
    single_active_consumer = group.get("settings", {}).get(
        "single_active_consumer", False
    )

    qspecs = [
        QueueSpec(
            name=name,
            vhost=vhost,
            arguments={
                "x-queue-type": qtype,
                **(
                    {"x-single-active-consumer": single_active_consumer}
                    if qtype != "stream"
                    else {}
                ),
                **(
                    {
                        "x-dead-letter-exchange": dlq_exchange,
                        "x-dead-letter-routing-key": dlq_pattern.format(name=name),
                    }
                    if dlq_pattern
                    else {}
                ),
            },
            auto_delete=False,
            durable=True,
        )
        for name in group["names"]
    ]

    # Add dead-letter queues within the default exchange with the "dlq." prefix
    if dlq_create and dlq_pattern:
        qspecs.extend(
            [
                QueueSpec(
                    name=dlq_pattern.format(name=name),
                    vhost=vhost,
                    arguments={
                        "x-queue-type": qtype,
                    },
                    auto_delete=False,
                    durable=True,
                )
                for name in group["names"]
            ]
        )
    return qspecs


def get_exchange_specs(exchanges: dict) -> List[ExchangeSpec]:
    return [
        ExchangeSpec(
            **exchange,
        )
        for exchange in exchanges
    ]


def get_exchange_specs_for_group(group: dict) -> List[ExchangeSpec]:
    vhost = group.get("vhost", "/")
    if group.get("settings", {}).get("broadcast"):
        etype = "fanout"
    else:
        etype = group.get("settings", {}).get("exchanges", {}).get("type", "direct")

    return [
        ExchangeSpec(
            arguments={},
            auto_delete=False,
            durable=True,
            name=name,
            type=etype,
            vhost=vhost,
        )
        for name in group["names"]
    ]


def _configure_policies(api, policies: List[Dict[str, Any]]):
    existing_policies = {
        (policy.vhost, policy.name): policy for policy in api.policies()
    }
    known_policies: set[Tuple[str, str]] = set()

    for policy in policies:
        policy_id = (policy["vhost"], policy["name"])
        if policy_id in known_policies:
            raise ValueError(f"Configuration defines duplicate policy {policy_id}")
        known_policies.add(policy_id)

        policy_spec = PolicySpec(
            vhost=policy.get("vhost", "/"),
            name=policy["name"],
            pattern=policy.get("pattern", "^amq."),
            definition=policy.get("definition", {}),
            priority=policy.get("priority", 0),
            apply_to=policy.get("apply-to", "queues"),
        )

        if policy_id in existing_policies:
            if existing_policies[policy_id] == policy_spec:
                continue
            logger.info(f"Updating policy: {policy_spec}")
        else:
            logger.info(f"Creating policy: {policy_spec}")
        api.set_policy(policy_spec)

    for policy_id in set(existing_policies) - known_policies:
        logger.info(f"Removing undefined policy {policy_id}")
        api.clear_policy(vhost=policy_id[0], name=policy_id[1])


def _configure_queues(api, queues: List[QueueSpec]):
    existing_queues = {(q.vhost, q.name): q for q in api.queues()}
    known_queues: set[Tuple[str, str]] = set()

    for queue_spec in queues:
        queue_id = (queue_spec.vhost, queue_spec.name)
        if queue_id in known_queues:
            raise ValueError(f"Configuration defines duplicate queue {queue_id}")
        known_queues.add(queue_id)

        if queue_id in existing_queues:
            equivalent_definition = all(
                getattr(existing_queues[queue_id], key) == value
                for key, value in queue_spec
            )
            if equivalent_definition:
                continue
            logger.info(f"Updating queue: {queue_spec}")
        else:
            logger.info(f"Creating queue: {queue_spec}")
        api.queue_declare(queue_spec)

    for queue_id in set(existing_queues) - known_queues:
        if (
            existing_queues[queue_id].name == ""
            or existing_queues[queue_id].name.startswith("amq.")
            or existing_queues[queue_id].auto_delete
            or existing_queues[queue_id].exclusive
        ):
            # Leave temporary queues alone
            continue
        logger.info(f"Removing undefined queue {queue_id}")
        api.queue_delete(vhost=queue_id[0], name=queue_id[1])


def _configure_users(api, rabbitmq_user_config_area: Path):
    existing_users = {user.name: user for user in api.users()}

    planned_users: Dict[str, Path] = {}
    for config_file in rabbitmq_user_config_area.glob("**/*.ini"):
        try:
            config = configparser.ConfigParser()
            config.read(config_file)
            username = config["rabbitmq"]["username"]
            password = config["rabbitmq"]["password"]
            assert username, "Configuration file does not contain a username"
            assert password, "Configuration file does not specify a password"
            tags = config["rabbitmq"].get("tags", "").split(",")
        except Exception:
            raise ValueError(f"Could not parse configuration file {config_file}")
        if username in planned_users:
            raise ValueError(
                f"Configuration file {config_file} declares user {username}, who was previously declared in {planned_users[username]}"
            )
        planned_users[username] = config_file

        if username in existing_users:
            hashed_password = hash_password(
                password, salt=existing_users[username].password_hash
            )
            if existing_users[username].password_hash == hashed_password and set(
                existing_users[username].tags
            ) == set(tags):
                continue
            logger.info(
                f"Updating user {username} due to password/tag mismatch (tags={tags})"
            )
        else:
            hashed_password = hash_password(password)
            logger.info(
                f"Creating user {username} not defined on the server (tags={tags})"
            )
        api.user_put(
            UserSpec(
                name=username,
                password_hash=hashed_password,
                hashing_algorithm="rabbit_password_hashing_sha256",
                tags=tags,
            )
        )

    for user in set(existing_users) - set(planned_users):
        logger.info(f"Removing user {user} not defined in the configuration")
        api.user_delete(name=user)


def run():
    logging.basicConfig(level=logging.INFO, format="%(message)s")
    zc = zocalo.configuration.from_file()
    zc.activate()
    parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument("configuration", help="RabbitMQ configuration yaml file")
    parser.add_argument(
        "--user-config",
        action="store",
        dest="user_config",
        type=Path,
        help="Path to directory containing *.ini files containing RabbitMQ\n"
        "    user credentials in the form:\n"
        "    [rabbitmq]\n"
        "    username = user\n"
        "    password = letmein\n"
        "    tags = comma,separated,tags",
    )
    zc.add_command_line_options(parser)
    args = parser.parse_args()

    api = RabbitMQAPI.from_zocalo_configuration(zc)

    with open(args.configuration) as in_file:
        yaml_data = yaml.safe_load(in_file)

    try:

        if args.user_config:
            _configure_users(api, args.user_config)

        # configure vhosts
        if vhost_specs := get_vhost_specs(yaml_data.get("vhosts", [])):
            current_vhosts_excluding_default = [
                vhost for vhost in api.vhosts() if vhost.name != "/"
            ]
            update_config(api, vhost_specs, current_vhosts_excluding_default)

        # configure permissions
        if permission_specs := get_permission_specs(yaml_data.get("permissions", [])):
            update_config(api, permission_specs, api.permissions())

        # configure policies
        _configure_policies(api, yaml_data["policies"])

        queue_specs = []
        exchange_specs = get_exchange_specs(yaml_data["exchanges"])
        binding_specs = get_binding_specs(yaml_data.get("bindings", []))
        for group in yaml_data["groups"]:
            if group.get("settings", {}).get("broadcast"):
                exchange_specs.extend(get_exchange_specs_for_group(group))
            else:
                queue_specs.extend(get_queue_specs(group))
                binding_specs.extend(get_binding_specs_for_group(group))

        _configure_queues(api, queue_specs)
        update_config(api, exchange_specs, api.exchanges())
        permanent_bindings = []
        # don't remove bindings to temporary queues
        queues = api.queues()
        for b in api.bindings():
            try:
                q = [
                    qu
                    for qu in queues
                    if qu.vhost == b.vhost and qu.name == b.destination
                ][0]
            except IndexError:
                logger.warning(f"No matching queue found binding {b}")
            else:
                if q.auto_delete or q.exclusive:
                    continue
                permanent_bindings.append(b)
        update_config(api, binding_specs, permanent_bindings)
    except requests.exceptions.HTTPError as e:
        logger.error(e)


if __name__ == "__main__":
    run()



from __future__ import annotations

from marshmallow import fields

from zocalo.configuration import PluginSchema


class JMX:
    class Schema(PluginSchema):
        host = fields.Str(required=True)
        port = fields.Int(required=True)
        base_url = fields.Str(required=True)
        username = fields.Str(required=True)
        password = fields.Str(required=True)

    @staticmethod
    def activate(configuration):
        return configuration



"""Top-level package for Zocalo."""

from __future__ import annotations

import logging

__author__ = "Diamond Light Source - Data Analysis Group"
__email__ = "dataanalysis@diamond.ac.uk"
__version__ = "0.25.0"

logging.getLogger("zocalo").addHandler(logging.NullHandler())


class ConfigurationError(Exception):
    pass



#
# zocalo.wrap
#   Wraps a command so that its status can be tracked in zocalo
#

from __future__ import annotations

import argparse
import faulthandler
import json
import logging
import signal
import sys

import pkg_resources
import workflows.recipe.wrapper
import workflows.services.common_service
import workflows.transport
import workflows.util

import zocalo.configuration.argparse
import zocalo.util
import zocalo.wrapper


def _enable_faulthandler():
    """Display a traceback on crashing with non-Python errors, such as
    segmentation faults, and when the process is signalled with SIGUSR2
    (not available on Windows)"""
    # Ignore errors during setup; SIGUSR2 not available on Windows, and
    # the attached STDERR might not support what faulthandler wants
    try:
        faulthandler.enable()
        faulthandler.register(signal.SIGUSR2)
    except Exception:
        pass


def run():
    zc = zocalo.configuration.from_file()
    zc.activate()

    known_wrappers = {
        e.name: e.load for e in pkg_resources.iter_entry_points("zocalo.wrappers")
    }

    # Set up parser
    parser = argparse.ArgumentParser(usage="zocalo.wrap [options]")
    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)

    parser.add_argument(
        "--wrap",
        action="store",
        dest="wrapper",
        metavar="WRAP",
        default=None,
        choices=list(known_wrappers),
        help="Object to be wrapped (valid choices: %s)" % ", ".join(known_wrappers),
    )
    parser.add_argument(
        "--recipewrapper",
        action="store",
        dest="recipewrapper",
        metavar="RW",
        default=None,
        help="A serialized recipe wrapper file for downstream communication",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Increase output verbosity",
    )

    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)

    # Parse command line arguments
    args = parser.parse_args()

    # Always enable logging to console
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    logging.getLogger().addHandler(console)
    if args.verbose:
        console.setLevel(logging.DEBUG)

    if zc.logging:
        zc.logging.verbosity = args.verbose
    else:
        logging.getLogger("workflows").setLevel(logging.INFO)
        logging.getLogger("zocalo").setLevel(logging.INFO)
        logging.getLogger().setLevel(logging.WARN)

    _enable_faulthandler()
    log = logging.getLogger("zocalo.wrap")

    # Instantiate specific wrapper
    if not args.wrapper:
        sys.exit("A wrapper object must be specified.")

    log.info(
        "Starting wrapper for %s with recipewrapper file %s",
        args.wrapper,
        args.recipewrapper,
    )

    # Connect to transport and start sending notifications
    transport = workflows.transport.lookup(args.transport)()
    transport.connect()
    st = zocalo.wrapper.StatusNotifications(transport.broadcast_status, args.wrapper)
    for field, value in zocalo.util.extended_status_dictionary().items():
        st.set_static_status_field(field, value)

    environment = {"config": zc}

    # Instantiate chosen wrapper
    instance = known_wrappers[args.wrapper]()(environment=environment)
    instance.status_thread = st

    # If specified, read in a serialized recipewrapper
    if args.recipewrapper:
        with open(args.recipewrapper) as fh:
            recwrap = workflows.recipe.wrapper.RecipeWrapper(
                message=json.load(fh), transport=transport
            )
        instance.set_recipe_wrapper(recwrap)

        if recwrap.recipe_step.get("wrapper", {}).get("task_information"):
            # If the recipe contains an extra task_information field then add this to the status display
            st.taskname += (
                " (" + str(recwrap.recipe_step["wrapper"]["task_information"]) + ")"
            )

    instance.prepare("Starting processing")

    st.set_status(workflows.services.common_service.Status.PROCESSING)
    log.info("Setup complete, starting processing")

    try:
        if instance.run():
            log.info("successfully finished processing")
            instance.success("Finished processing")
        else:
            log.info("processing failed")
            instance.failure("Processing failed")
        st.set_status(workflows.services.common_service.Status.END)
    except KeyboardInterrupt:
        log.info("Shutdown via Ctrl+C")
        st.set_status(workflows.services.common_service.Status.END)
    except Exception as e:
        log.error(str(e), exc_info=True)
        instance.failure(e)
        st.set_status(workflows.services.common_service.Status.ERROR)

    instance.done("Finished processing")

    st.shutdown()
    st.join()
    log.debug("Terminating")
    transport.disconnect()



from __future__ import annotations

import pytest

import zocalo.configuration

sample_configuration = """
version: 1

example:
  plugin: jmx
  host: localhost
  port: 8080
  base_url: /somewhere
  username: admin
  password: admin

environments:
  live:
    - example
"""


def test_plugin_makes_jmx_config_available():
    zc = zocalo.configuration.from_string(sample_configuration)
    assert zc.jmx is None
    zc.activate_environment("live")
    assert isinstance(zc.jmx, dict)
    assert zc.jmx["host"] == "localhost"


def test_invalid_configuration_is_rejected():
    with pytest.raises(zocalo.ConfigurationError, match="integer"):
        zocalo.configuration.from_string(sample_configuration.replace("8080", "banana"))

    with pytest.raises(zocalo.ConfigurationError, match="username"):
        zocalo.configuration.from_string(
            """
            version: 1
            failure:
              plugin: jmx
            """
        )



from __future__ import annotations

import json
import threading
from pathlib import Path

import workflows.recipe
from workflows.services.common_service import CommonService


class JSONLines(CommonService):
    """Write received messages into a JSONLines file on disk"""

    # Human readable service name
    _service_name = "JSON Lines"

    # Logger name
    _logger_name = "zocalo.service.jsonlines"

    def initializing(self):

        self._register_idle(1, self.process_messages)
        workflows.recipe.wrap_subscribe(
            self._transport,
            "jsonlines",
            self.receive_msg,
            acknowledgement=True,
            exclusive=True,
            log_extender=self.extend_log,
            prefetch_count=100,
        )
        self._lock = threading.Lock()
        self._data: dict[Path, list[tuple(dict, dict)]] = {}

    def receive_msg(
        self, rw: workflows.recipe.RecipeWrapper, header: dict, message: dict
    ):
        output_filename = rw.recipe_step["parameters"]["output_filename"]
        if not output_filename:
            self.log.error("Received message contains no output_filename")
            rw.transport.nack(header)
            return
        output_filename = Path(output_filename)
        exclude = set(rw.recipe_step["parameters"].get("exclude", []))
        include = set(rw.recipe_step["parameters"].get("include", []))
        filtered_message = message
        if include:
            filtered_message = {
                k: v for k, v in filtered_message.items() if k in include
            }
        if exclude:
            filtered_message = {
                k: v for k, v in filtered_message.items() if k not in exclude
            }

        with self._lock:
            self._data.setdefault(output_filename, [])
            self._data[output_filename].append((header, filtered_message))

    def process_messages(self):
        with self._lock:
            for output_filename in list(self._data):
                grouped_data = self._data[output_filename]
                self.log.info(
                    f"Writing {len(grouped_data)} messages to {output_filename}"
                )
                try:
                    output_filename = Path(output_filename)
                    output_filename.parent.mkdir(exist_ok=True, parents=True)
                    with output_filename.open(mode="a") as fh:
                        for _, message in grouped_data:
                            fh.write(json.dumps(message) + "\n")
                except Exception as e:
                    self.log.error(
                        f"Uncaught exception {e!r} writing messages to {output_filename}",
                        exc_info=True,
                    )
                    for header, _ in grouped_data:
                        self.transport.nack(header)
                    return
                else:
                    for header, _ in grouped_data:
                        self.transport.ack(header)

                # delete this data now we've processed it
                del self._data[output_filename]



#
# zocalo.dlq_reinject
#   Take a dead letter queue message from a file and send it back to its queue
#   for a retry.
#
from __future__ import annotations

import argparse
import json
import os
import re
import select
import sys
import time
from pprint import pprint

import workflows.transport

import zocalo.configuration
from zocalo.util.rabbitmq import RabbitMQAPI


def run() -> None:
    zc = zocalo.configuration.from_file()
    zc.activate()
    parser = argparse.ArgumentParser(
        usage="zocalo.dlq_reinject [options] file [file [..]]"
    )

    parser.add_argument("-?", action="help", help=argparse.SUPPRESS)
    parser.add_argument(
        "-r",
        "--remove",
        action="store_true",
        default=False,
        dest="remove",
        help="Delete file on successful reinjection",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        dest="verbose",
        help="Show message contents",
    )
    parser.add_argument(
        "-d",
        "--destination",
        action="store",
        default=None,
        dest="destination_override",
        help="Reinject messages to a different destination. Any name given must include the stomp prefix.",
    )
    parser.add_argument(
        "-w",
        "--wait",
        default=None,
        dest="wait",
        help="Wait this many seconds between reinjections",
    )
    parser.add_argument(
        "files", nargs="*", help="File(s) containing DLQ messages to be reinjected"
    )

    zc.add_command_line_options(parser)
    workflows.transport.add_command_line_options(parser, transport_argument=True)
    args = parser.parse_args()
    transport = workflows.transport.lookup(args.transport)()

    stdin = []
    try:
        stdin_fileno = getattr(sys.stdin, "fileno", lambda: None)()
    except Exception:
        stdin_fileno = None
    if isinstance(stdin_fileno, int) and select.select([sys.stdin], [], [], 0.0)[0]:
        dlq_purge_filename_format = re.compile(r"^  \/")
        while True:
            line = sys.stdin.readline()
            if not line:
                break
            if dlq_purge_filename_format.match(line):
                stdin.append(line.strip())
        print(f"{len(stdin)} filenames read from stdin")

    if not stdin and not args.files:
        sys.exit("No DLQ message files given.")

    transport.connect()

    first = True
    for dlqfile in args.files + stdin:
        if not os.path.exists(dlqfile):
            print(f"Ignoring missing file {dlqfile}")
            continue
        if not first and args.wait:
            time.sleep(float(args.wait))
        first = False
        with open(dlqfile) as fh:
            dlqmsg = json.load(fh)
        print(f"Parsing message from {dlqfile}")
        if (
            not isinstance(dlqmsg, dict)
            or not dlqmsg.get("header")
            or not dlqmsg.get("message")
        ):
            sys.exit("File is not a valid DLQ message.")
        if args.verbose:
            pprint(dlqmsg)

        if args.transport == "StompTransport":
            destination = (
                dlqmsg["header"]
                .get("original-destination", dlqmsg["header"]["destination"])
                .split("/", 2)
            )
            if destination[1] == "queue":
                print("sending...")
                send_function = transport.send
            elif destination[1] == "topic":
                print("broadcasting...")
                send_function = transport.broadcast
            else:
                sys.exit("Cannot process message, unknown message mechanism")
            if args.destination_override:
                destination[2] = args.destination_override
            header = dlqmsg["header"]
            for drop_field in (
                "content-length",
                "destination",
                "expires",
                "message-id",
                "original-destination",
                "originalExpiration",
                "subscription",
                "timestamp",
                "redelivered",
            ):
                if drop_field in header:
                    del header[drop_field]
            header["dlq-reinjected"] = "True"
            send_function(
                destination[2], dlqmsg["message"], headers=header, ignore_namespace=True
            )
        elif args.transport == "PikaTransport":
            print("pika transport detected")
            header = dlqmsg["header"]
            header["dlq-reinjected"] = "True"
            exchange = header.get("x-death", [{}])[0].get("exchange")
            if exchange:
                rmqapi = RabbitMQAPI.from_zocalo_configuration(zc)
                exchange_info = rmqapi.get("queues").json()
                for exch in exchange_info:
                    if exch["name"] == exchange:
                        if exch["type"] == "fanout":
                            transport.broadcast(
                                args.destination_override or destination,
                                dlqmsg["message"],
                                headers=_rabbit_prepare_header(header),
                            )
            else:
                destination = header.get("x-death", [{}])[0].get("queue")
                transport.send(
                    args.destination_override or destination,
                    dlqmsg["message"],
                    headers=_rabbit_prepare_header(header),
                )
        if args.remove:
            os.remove(dlqfile)
        print("Done.\n")

    transport.disconnect()


def _rabbit_prepare_header(header: dict) -> dict:
    drop = {
        "message-id",
        "routing_key",
        "redelivered",
        "exchange",
        "consumer_tag",
        "delivery_mode",
    }
    return {k: str(v) for k, v in header.items() if k not in drop}



from __future__ import annotations

import email.message
import pprint
import smtplib

import workflows.recipe
from workflows.services.common_service import CommonService

import zocalo.configuration


class _SafeDict(dict):
    """A dictionary that returns undefined keys as {keyname}.
    This can be used to selectively replace variables in datastructures."""

    def __missing__(self, key):
        return "{" + key + "}"


class Mailer(CommonService):
    """A service that generates emails from messages."""

    # Human readable service name
    _service_name = "Mail Notifications"

    # Logger name
    _logger_name = "zocalo.services.mailer"

    def initializing(self):
        """Subscribe to the Mail notification queue.
        Received messages must be acknowledged."""
        self.log.debug("Mail notifications starting")

        if not self.config:
            raise zocalo.ConfigurationError("No Zocalo configuration loaded")
        if not self.config.smtp:
            raise zocalo.ConfigurationError(
                "There are no SMTP settings configured in your environment"
            )

        workflows.recipe.wrap_subscribe(
            self._transport,
            "mailnotification",
            self.receive_msg,
            acknowledgement=True,
            log_extender=self.extend_log,
            allow_non_recipe_messages=True,
        )

    @staticmethod
    def listify(recipients):
        if isinstance(recipients, list):
            return recipients
        elif isinstance(recipients, tuple):
            return list(recipients)
        else:
            return [recipients]

    def receive_msg(self, rw, header, message):
        """Do some mail notification."""

        self.log.info(f"{message=}")

        if rw:
            parameters = rw.recipe_step["parameters"]
            content = None
        else:
            # Incoming message is not a recipe message. Simple messages can be valid
            if (
                not isinstance(message, dict)
                or not message.get("parameters")
                or not message.get("content")
            ):
                self.log.warning("Rejected invalid simple message")
                self._transport.nack(header)
                return

            parameters = message["parameters"]
            content = message["content"]

        recipients = parameters.get("recipients", parameters.get("recipient"))
        if not recipients:
            self.log.warning("No recipients set for message")
            self._transport.nack(header)
            return
        if isinstance(recipients, dict):
            if "select" not in recipients:
                self.log.warning(
                    "Recipients dictionary must have key 'select' to select relevant group"
                )
                self._transport.nack(header)
                return
            selected_recipients = self.listify(recipients.get(recipients["select"], []))
            if recipients.get("all"):
                all_recipients = self.listify(recipients["all"])
            recipients = sorted(set(selected_recipients) | set(all_recipients))
            if not recipients:
                self.log.warning("No selected recipients for message")
                self._transport.nack(header)
                return
        else:
            recipients = self.listify(recipients)

        sender = parameters.get("from", self.config.smtp["from"])

        subject = parameters.get("subject", "mail notification via zocalo")

        content = parameters.get("content", content)

        if not content:
            self.log.warning("Message has no content")
            self._transport.nack(header)
            return
        if isinstance(content, list):
            content = "".join(content)
        if isinstance(message, list):
            pprint_message = "\n".join(message)
        else:
            pprint_message = pprint.pformat(message)

        content = content.format_map(
            _SafeDict(payload=message, pprint_payload=pprint_message)
        )

        self.log.info("Sending mail notification %r to %r", subject, recipients)

        # Accept message before sending mail. While this means we do not guarantee
        # message delivery it also means if the service crashes after delivery we
        # will not re-deliver the message inifinitely many times.
        self._transport.ack(header)

        try:
            msg = email.message.EmailMessage()
            msg["Subject"] = subject
            msg["To"] = recipients
            msg["From"] = sender
            msg.set_content(content)
            with smtplib.SMTP(
                host=self.config.smtp["host"], port=self.config.smtp["port"], timeout=60
            ) as s:
                s.send_message(msg)
        except TimeoutError as e:
            self.log.error(
                f"Message delivery failed with timeout: {e}",
            )
        except Exception as e:
            self.log.error(
                f"Message delivery failed with error {e}",
            )
        else:
            self.log.debug("Message sent successfully")



from __future__ import annotations

from unittest import mock

import pytest

import zocalo.configuration

sample_configuration = """
version: 1

graylog:
  plugin: graylog
  protocol: UDP
  host: localhost
  port: 12201

environments:
  live:
    - graylog
"""


@mock.patch("zocalo.configuration.plugin_graylog.graypy")
@mock.patch("zocalo.configuration.plugin_graylog.graypy.handler")
@mock.patch("zocalo.configuration.plugin_graylog.logging")
def test_plugin_sets_up_logging(logging, handler, graypy):
    zc = zocalo.configuration.from_string(sample_configuration)
    graypy.GELFUDPHandler.assert_not_called()
    logging.getLogger.return_value.addHandler.assert_not_called()

    zc.activate_environment("live")
    graypy.GELFUDPHandler.assert_called_once_with("127.0.0.1", 12201, level_names=True)
    relevant_handler = graypy.GELFUDPHandler.return_value
    logging.getLogger.return_value.addHandler.assert_called_once_with(relevant_handler)
    assert zc.graylog == relevant_handler


def test_invalid_configuration_is_rejected():
    with pytest.raises(zocalo.ConfigurationError, match="TCP"):
        zocalo.configuration.from_string(sample_configuration.replace("UDP", "banana"))

    with pytest.raises(zocalo.ConfigurationError, match="host"):
        zocalo.configuration.from_string(
            """
            version: 1
            graylog:
              plugin: graylog
            """
        )



#!/usr/bin/env python

from __future__ import annotations

from setuptools import setup

_rst_section = "\n--------"

with open("README.rst") as readme_file:
    readme = readme_file.read()

    # Only take text up to the first section
    if _rst_section in readme:
        readme = readme.split(_rst_section)[0]
        readme = "\n".join(readme.split("\n")[:-1])

with open("HISTORY.rst") as history_file:
    history = history_file.read()

    # Only take text from up to 6 sections
    _history_split = history.split(_rst_section)
    if len(_history_split) > 7:
        history = _rst_section.join(_history_split[:7])
        history = "\n".join(history.split("\n")[:-1])

setup(
    long_description=readme + "\n\n" + history,
)



from __future__ import annotations

import logging
import sys

import workflows.contrib.start_service

import zocalo.configuration.argparse
import zocalo.util


def start_service():
    ServiceStarter().run(
        program_name="zocalo.service",
        version=zocalo.__version__,
        transport_command_channel="command",
    )


class ServiceStarter(workflows.contrib.start_service.ServiceStarter):
    """Starts a workflow service"""

    __frontendref = None

    def setup_logging(self):
        """Initialize common logging framework. Everything is logged to central
        graylog server. Depending on setting messages of DEBUG or INFO and higher
        go to console."""

        # Always enable logging to console
        try:
            from dlstbx.util.colorstreamhandler import ColorStreamHandler

            self.console = ColorStreamHandler()
        except ImportError:
            self.console = logging.StreamHandler()
        self.console.setLevel(logging.INFO)
        logging.getLogger().addHandler(self.console)

        if not self._zc.logging:
            logging.getLogger().setLevel(logging.WARN)
            logging.getLogger("workflows").setLevel(logging.INFO)
            logging.getLogger("zocalo").setLevel(logging.DEBUG)
            logging.getLogger("zocalo.service").setLevel(logging.DEBUG)

        self.log = logging.getLogger("zocalo.service")

    def __init__(self):
        # load configuration and initialize logging
        self._zc = zocalo.configuration.from_file()
        envs = self._zc.activate()
        self.use_live_infrastructure = ("live" in envs) or (
            "default" in envs
        )  # deprecated
        self.setup_logging()

        if (
            self._zc.storage
            and self._zc.storage.get("zocalo.default_transport")
            in workflows.transport.get_known_transports()
        ):
            workflows.transport.default_transport = self._zc.storage[
                "zocalo.default_transport"
            ]

    def on_parser_preparation(self, parser):
        parser.add_option(
            "-v",
            "--verbose",
            action="count",
            default=0,
            help="Increase output verbosity",
        )
        parser.add_option(
            "--tag",
            dest="tag",
            metavar="TAG",
            default=None,
            help="Individual tag related to this service instance",
        )
        parser.add_option(
            "-d",
            "--debug",
            action="store_true",
            dest="debug",
            default=False,
            help="Set debug log level for workflows",
        )
        parser.add_option(
            "-r",
            "--restart",
            action="store_true",
            dest="service_restart",
            default=False,
            help="Restart service on failure",
        )
        self._zc.add_command_line_options(parser)
        self.log.debug("Launching %r", sys.argv)

    def on_parsing(self, options, args):
        if options.verbose:
            self.console.setLevel(logging.DEBUG)
            if self._zc.logging:
                self._zc.logging.verbosity = options.verbose
        if options.debug:
            self.console.setLevel(logging.DEBUG)
            if not self._zc.logging:
                logging.getLogger("pika").setLevel(logging.INFO)
                logging.getLogger("stomp.py").setLevel(logging.DEBUG)
                logging.getLogger("workflows").setLevel(logging.DEBUG)
        self.options = options

    def before_frontend_construction(self, kwargs):
        kwargs["verbose_service"] = True
        kwargs["environment"] = kwargs.get("environment", {})
        kwargs["environment"]["live"] = self.use_live_infrastructure
        kwargs["environment"]["config"] = self._zc
        return kwargs

    def on_frontend_preparation(self, frontend):
        if self.options.service_restart:
            frontend.restart_service = True

        extended_status = zocalo.util.extended_status_dictionary()
        if self.options.tag:
            extended_status["tag"] = self.options.tag

        original_status_function = frontend.get_status

        def extend_status_wrapper():
            status = original_status_function()
            status.update(extended_status)
            return status

        frontend.get_status = extend_status_wrapper



from __future__ import annotations

import pytest

import zocalo.configuration

sample_configuration = """
version: 1

my-smtp:
  plugin: smtp
  host: localhost
  port: 8080
  from: foo@example.com

environments:
  live:
    - my-smtp
"""


def test_plugin_makes_smtp_config_available():
    zc = zocalo.configuration.from_string(sample_configuration)
    assert zc.smtp is None
    zc.activate_environment("live")
    assert isinstance(zc.smtp, dict)
    assert zc.smtp["host"] == "localhost"
    assert zc.smtp["port"] == 8080
    assert zc.smtp["from"] == "foo@example.com"


def test_invalid_configuration_is_rejected():
    with pytest.raises(zocalo.ConfigurationError, match="integer"):
        zocalo.configuration.from_string(sample_configuration.replace("8080", "banana"))

    with pytest.raises(zocalo.ConfigurationError, match="host"):
        zocalo.configuration.from_string(
            """
            version: 1
            failure:
              plugin: smtp
            """
        )



from __future__ import annotations

import json
import sys
import time
from datetime import datetime
from unittest import mock

import workflows.transport
from workflows.transport.common_transport import CommonTransport

from zocalo.cli.dlq_reinject import run


def gen_header_activemq(i):
    return {
        "content-length": "2489",
        "expires": "0",
        "destination": "/queue/zocalo.garbage.per_image_analysis",
        "subscription": "1",
        "message-id": f"ID:foo.bar.com-{i}",
        "timestamp": "1633102156582",
    }


def gen_header_rabbitmq(i, use_datetime=True):
    tstamp = 1633962302 + 30 * i
    if use_datetime:
        tstamp = datetime.fromtimestamp(tstamp)
    return {
        "x-death": [{"time": tstamp, "queue": "garbage.per_image_analysis"}],
        "message-id": f"ID:foo.bar.com-{i}",
    }


def test_dlq_reinject_activemq(mocker, tmp_path):
    mocked_transport = mocker.MagicMock(CommonTransport)
    mocker.patch.object(workflows.transport, "lookup", return_value=mocked_transport)

    dlq_path = tmp_path / "DLQ"
    dlq_path.mkdir()

    for i in range(10):
        with open(dlq_path / f"msg_{i}", "w") as f:
            dlqmsg = {
                "exported": {
                    "date": time.strftime("%Y-%m-%d"),
                    "time": time.strftime("%H:%M:%S"),
                },
                "header": gen_header_activemq(i),
                "message": {
                    "foo": f"{i}",
                },
            }
            json.dump(dlqmsg, f)

    testargs = ["prog", "--transport", "StompTransport"] + [
        str(dlq_path / f"msg_{i}") for i in range(10)
    ]
    with mock.patch.object(sys, "argv", testargs):
        run()

    mocked_transport().send.assert_has_calls(
        [
            mock.call(
                "zocalo.garbage.per_image_analysis",
                {"foo": f"{i}"},
                headers=mock.ANY,
                ignore_namespace=True,
            )
            for i in range(10)
        ]
    )


def test_dlq_reinject_rabbitmq(mocker, tmp_path):
    mocked_transport = mocker.MagicMock(CommonTransport)
    mocker.patch.object(workflows.transport, "lookup", return_value=mocked_transport)

    dlq_path = tmp_path / "DLQ"
    dlq_path.mkdir()

    for i in range(10):
        with open(dlq_path / f"msg_{i}", "w") as f:
            dlqmsg = {
                "exported": {
                    "date": time.strftime("%Y-%m-%d"),
                    "time": time.strftime("%H:%M:%S"),
                },
                "header": gen_header_rabbitmq(i, use_datetime=False),
                "message": {
                    "foo": f"{i}",
                },
            }
            json.dump(dlqmsg, f)

    testargs = ["prog", "--transport", "PikaTransport"] + [
        str(dlq_path / f"msg_{i}") for i in range(10)
    ]
    with mock.patch.object(sys, "argv", testargs):
        run()

    mocked_transport().send.assert_has_calls(
        [
            mock.call(
                "garbage.per_image_analysis",
                {"foo": f"{i}"},
                headers=mock.ANY,
            )
            for i in range(10)
        ]
    )
